import os
import tempfile
from pathlib import Path
from typing import Any, List, Tuple

from .mesh import TetraMesh, TriangleMesh
from .region_markers import RegionMarker


def write_smesh(filename: os.PathLike,
                mesh: TriangleMesh,
                region_markers: List[RegionMarker] = None):
    """Save a mesh to a `.smesh` format (Tetgen). http://wias-
    berlin.de/software/tetgen/1.5/doc/manual/manual006.html#ff_smesh.

    Parameters
    ----------
    filename : os.PathLike
        Filename to save the data to.
    mesh : TriangleMesh
        Mesh data to be saved.
    region_markers : List[RegionMarker], optional
        Override region markers from input mesh.
    """
    if region_markers is None:
        region_markers = mesh.region_markers

    path = Path(filename)
    with path.open('w') as f:
        n_nodes, n_dim = mesh.points.shape
        n_attrs = 0
        node_markers = 0

        print(f'{n_nodes} {n_dim} {n_attrs} {node_markers}', file=f)

        node_fmt = '{:4d}' + ' {:8.2f}' * n_dim

        for i, node in enumerate(mesh.points):
            print(node_fmt.format(i + 1, *node), file=f)

        n_facets, n_corners = mesh.cells.shape
        facet_markers = 0

        print(f'{n_facets} {facet_markers}', file=f)

        facet_fmt = '{:4d}' + ' {:8d}' * n_corners

        for facet in mesh.cells + 1:  # tetgen uses 1-indexing
            print(facet_fmt.format(n_corners, *facet), file=f)

        # TODO, store holes in TriangleMesh?
        n_holes = 0
        hole_dim = 3

        print(f'{n_holes}', file=f)
        holes: Tuple[Any, ...] = ()

        hole_fmt = '{:4d}' + ' {:8.2f}' * hole_dim

        for i, hole in enumerate(holes):
            print(hole_fmt.format(i + 1, *hole), file=f)

        n_regions = len(region_markers)
        region_dim = 3

        print(f'{n_regions}', file=f)

        region_fmt = '{:4d}' + ' {:8.2f}' * region_dim + ' {label:8}'

        for i, marker in enumerate(region_markers):
            print(region_fmt.format(i + 1,
                                    *marker.coordinates,
                                    label=marker.label),
                  file=f)


def call_tetgen(fname: os.PathLike, opts: str = '-pAq1.2'):
    """Call tetgen via subprocess.

    Parameters
    ----------
    fname : os.PathLike
        Location of tetgen input file ('.smesh')
    opts : str
        Command-line options passed to `tetgen`.

        More info:
        http://wias-berlin.de/software/tetgen/1.5/doc/manual/manual005.html

        Some useful flags:

        -A: Assigns attributes to tetrahedra in different regions.
        -p: Tetrahedralizes a piecewise linear complex (PLC).
        -q: Refines mesh (to improve mesh quality).
        -a: Applies a maximum tetrahedron volume constraint.
    """
    import subprocess as sp
    sp.run(['tetgen', opts, fname])


def tetrahedralize(mesh: TriangleMesh, opts: str = '-pAq1.2') -> TetraMesh:
    """Tetrahedralize a surface mesh.

    Parameters
    ----------
    mesh : TriangleMesh
        Input contour mesh
    opts : str, optional
        Command-line options passed to `tetgen`.

        More info:
        http://wias-berlin.de/software/tetgen/1.5/doc/manual/manual005.html

        Some useful flags:

        -A: Assigns attributes to tetrahedra in different regions.
        -p: Tetrahedralizes a piecewise linear complex (PLC).
        -q: Refines mesh (to improve mesh quality).
        -a: Applies a maximum tetrahedron volume constraint.

    Returns
    -------
    TetraMesh
        Tetrahedralized mesh.
    """
    with tempfile.TemporaryDirectory() as tmp:
        path = Path(tmp, 'nanomesh.smesh')
        write_smesh(path, mesh)
        call_tetgen(path, opts)
        ele_path = path.with_suffix('.1.ele')
        return TetraMesh.read(ele_path)



from dataclasses import dataclass
from typing import Optional, Tuple, Union

import numpy.typing as npt


@dataclass
class RegionMarker:
    label: int
    coordinates: npt.NDArray
    name: Optional[str] = None


RegionMarkerLike = Union[RegionMarker, Tuple[int, npt.NDArray]]



import operator
import os
from typing import Callable, Union

import numpy as np


def normalize_values(image: np.ndarray):
    """Rescale values to 0.0 to 1.0.

    Parameters
    ----------
    image : (m, n) np.ndarray
        Input image.

    Returns
    -------
    out : (m, n) np.ndarray
        Normalized image
    """
    out = (image - image.min()) / (image.max() - image.min())
    return out


class BaseImage:
    """Data class for image data.

    Parameters
    ----------
    image : np.array
        N-dimensional numpy array containing image data.
    """
    def __init__(self, image: np.ndarray):
        self.image = image

    def __repr__(self):
        return f'{self.__class__.__name__}(shape={self.image.shape})'

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return np.all(other.image == self.image)
        elif isinstance(other, np.ndarray):
            return np.all(other == self.image)
        else:
            return False

    def __gt__(self, other):
        return self._compare(other, op=operator.gt)

    def __lt__(self, other):
        return self._compare(other, op=operator.lt)

    def __ge__(self, other):
        return self._compare(other, op=operator.ge)

    def __le__(self, other):
        return self._compare(other, op=operator.le)

    def _compare(self, other, *, op: Callable):
        """Helper function to implement overload functions.

        Parameters
        ----------
        other :
            Other instance, can be a numpy array.
        op : callable
            Operator (see operator module).

        Returns
        -------
        BaseImage
            Image with boolean data.
        """
        this = self.image
        if isinstance(other, self.__class__):
            other = other.image
        return self.__class__(op(this, other))

    def to_sitk_image(self):
        """Return instance of `SimpleITK.Image` from `.image`."""
        import SimpleITK as sitk
        return sitk.GetImageFromArray(self.image)

    @classmethod
    def from_sitk_image(cls, sitk_image) -> 'BaseImage':
        """Return instance from `SimpleITK.Image`."""
        import SimpleITK as sitk
        image = sitk.GetArrayFromImage(sitk_image)
        return cls(image)

    def save(self, filename: os.PathLike):
        """Save the data. Supported filetypes: `.npy`.

        Parameters
        ----------
        filename : Pathlike
            Name of the file to save to.
        """
        np.save(filename, self.image)

    @classmethod
    def load(cls, filename: os.PathLike, **kwargs):
        """Load the data. Supported filetypes: `.npy`.

        Parameters
        ----------
        filename : Pathlike
            Name of the file to load.
        **kwargs : dict
            Extra keyword arguments passed to `np.load`.
        """
        image = np.load(filename, **kwargs)
        return cls(image)

    def apply(self, function: Callable, **kwargs):
        """Apply function to `.image` array. Return an instance of `BaseImage`
        if the result is of the same dimensions, otherwise return the result of
        the operation.

        Parameters
        ----------
        function : callable
            Function to apply to `self.image`.
        **kwargs
            Keyword arguments to pass to `function`.

        Returns
        -------
        BaseImage
            New instance of `BaseImage`.
        """
        ret = function(self.image, **kwargs)
        if isinstance(ret, np.ndarray) and (ret.ndim == self.image.ndim):
            return self.__class__(ret)

        return ret

    def gaussian(self, sigma: int = 1, **kwargs):
        """Apply Gaussian blur to image.

        Parameters
        ----------
        sigma : int
            Standard deviation for Gaussian kernel.
        **kwargs : dict
            Extra arguments passed to `skimage.filters.gaussian`.

        Returns
        -------
        BaseImage
            New instance of `BaseImage`.
        """
        from skimage import filters
        return self.apply(filters.gaussian, sigma=sigma, **kwargs)

    def digitize(self, bins: Union[list, tuple], **kwargs):
        """Digitize image.

        For more info see `numpy.digitize`.

        Parameters
        ----------
        bins : list, tuple
            List of bin values. Must be monotonic and one-dimensional.
        **kwargs : dict
            Extra arguments passed to `numpy.digitize`.

        Returns
        -------
        BaseImage
            New instance of `BaseImage`.
        """
        return self.apply(np.digitize, bins=bins, **kwargs)

    def normalize_values(self):
        """Rescale values to 0.0 to 1.0.

        Returns
        -------
        out : BaseImage
            Normalized image
        """
        return self.apply(normalize_values)

    def binary_digitize(self, threshold: Union[float, str] = None):
        """Convert into a binary image.

        Parameters
        ----------
        threshold : float, optional
            Threshold used for segmentation. If given as a string,
            apply corresponding theshold via `BaseImage.threshold`.
            Defaults to median.

        Returns
        -------
        BaseImage
            New instance of `BaseImage`.
        """
        if not threshold:
            threshold = np.median(self.image)
        if isinstance(threshold, str):
            threshold = self.threshold(threshold)
        return self.apply(np.digitize, bins=[threshold])

    def threshold(self, method: str = 'otsu', **kwargs) -> float:
        """Compute threshold value using given method.

        For more info, see:
            https://scikit-image.org/docs/dev/api/skimage.filters.html

        Parameters
        ----------
        method : str
            Thresholding method to use. Defaults to `otsu`.
        **kwargs : dict
            Extra keyword arguments passed to threshold method.

        Returns
        -------
        threshold : float
            Threshold value.
        """
        from skimage import filters
        dispatch_table = {
            'isodata': filters.threshold_isodata,
            'li': filters.threshold_li,
            'local': filters.threshold_local,
            'mean': filters.threshold_mean,
            'minimum': filters.threshold_minimum,
            'multiotsu': filters.threshold_multiotsu,
            'niblack': filters.threshold_niblack,
            'otsu': filters.threshold_otsu,
            'sauvola': filters.threshold_sauvola,
            'triangle': filters.threshold_triangle,
            'yen': filters.threshold_yen,
        }
        try:
            func = dispatch_table[method]
        except KeyError:
            raise KeyError(f'`method` must be one of {dispatch_table.keys()}')

        return self.apply(func, **kwargs)

    def fft(self) -> 'BaseImage':
        """Apply fourier transform to image.

        Returns
        -------
        BaseImage
            Real component of fourier transform with the zero-frequency
            component shifted to the center of the spectrum.
        """
        fourier = np.fft.fftn(self.image)
        shifted = np.abs(np.fft.fftshift(fourier))
        return self.__class__(shifted)



from dataclasses import dataclass
from typing import Tuple

import numpy as np


@dataclass
class BoundingBox:
    """Container for bounding box coordinates."""
    xmin: float
    xmax: float
    ymin: float
    ymax: float
    zmin: float
    zmax: float

    @classmethod
    def from_shape(cls, shape: tuple):
        """Generate bounding box from data shape."""
        xmin, ymin, zmin = 0, 0, 0
        xmax, ymax, zmax = np.array(shape) - 1
        return cls(
            xmin=xmin,
            ymin=ymin,
            zmin=zmin,
            xmax=xmax,
            ymax=ymax,
            zmax=zmax,
        )

    @property
    def dimensions(self) -> Tuple[float, float, float]:
        """Return dimensions of bounding box."""
        return (
            self.xmax - self.xmin,
            self.ymax - self.ymin,
            self.zmax - self.zmin,
        )

    @classmethod
    def from_points(cls, points: np.ndarray):
        """Generate bounding box from set of points or coordinates."""
        xmax, ymax, zmax = np.max(points, axis=0)
        xmin, ymin, zmin = np.min(points, axis=0)

        return cls(
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
            zmin=zmin,
            zmax=zmax,
        )

    def to_points(self) -> np.ndarray:
        """Return (m,3) array with corner points."""
        return np.array([
            [self.xmin, self.ymin, self.zmin],
            [self.xmin, self.ymin, self.zmax],
            [self.xmin, self.ymax, self.zmin],
            [self.xmin, self.ymax, self.zmax],
            [self.xmax, self.ymin, self.zmin],
            [self.xmax, self.ymin, self.zmax],
            [self.xmax, self.ymax, self.zmin],
            [self.xmax, self.ymax, self.zmax],
        ])

    @property
    def center(self) -> np.ndarray:
        """Return center of the bounding box."""
        return np.array((
            (self.xmin + self.xmax) / 2,
            (self.ymin + self.ymax) / 2,
            (self.zmin + self.zmax) / 2,
        ))



from __future__ import annotations

from types import MappingProxyType
from typing import TYPE_CHECKING, Dict, List, Sequence

import matplotlib.pyplot as plt
import meshio
import numpy as np
import pyvista as pv
import scipy

from .mpl.meshplot import _legend_with_triplot_fix
from .region_markers import RegionMarker, RegionMarkerLike

if TYPE_CHECKING:
    import open3d

    from .mesh_container import MeshContainer


class PruneZ0Mixin:
    def prune_z_0(self):
        """Drop third dimension (z) coordinates if present and all values are
        equal to 0 (within tolerance).

        For compatibility, sometimes a column with zeroes is added, for
        example when exporting to gmsh2.2 format. This method drops that
        column.
        """
        TOL = 1e-9

        is_3_dimensional = self.points.shape[1] == 3
        if not is_3_dimensional:
            return

        if not np.all(np.abs(self.points[:, 2]) < TOL):
            raise ValueError(
                'Coordinates in third dimension are not all equal to zero.')

        self.points = self.points[:, 0:2]


class BaseMesh:
    _cell_type: str = 'base'

    def __init__(self,
                 points: np.ndarray,
                 cells: np.ndarray,
                 fields: Dict[str, int] = None,
                 region_markers: List[RegionMarker] = None,
                 **cell_data):
        """Base class for meshes.

        Parameters
        ----------
        points : (m, n) np.ndarray[float]
            Array with points.
        cells : (i, j) np.ndarray[int]
            Index array describing the cells of the mesh.
        fields : Dict[str, int]:
            Mapping from field names to labels
        region_markers : List[RegionMarker], optional
            List of region markers used for assigning labels to regions.
            Defaults to an empty list.
        **cell_data
            Additional cell data. Argument must be a 1D numpy array
            matching the number of cells defined by `i`.
        """
        default_key = 'physical'
        if (not cell_data) or (default_key in cell_data):
            self.default_key = default_key
        else:
            self.default_key = list(cell_data.keys())[0]

        if not fields:
            fields = {}

        self.points = points
        self.cells = cells
        self.field_to_number = fields
        self.region_markers = [] if region_markers is None else region_markers
        self.cell_data = cell_data

    @property
    def number_to_field(self):
        """Mapping from numbers to fields, proxy to `.field_to_number`."""
        return MappingProxyType(
            {v: k
             for k, v in self.field_to_number.items()})

    def add_region_marker(self, region_marker: RegionMarkerLike):
        """Add marker to list of region markers.

        Parameters
        ----------
        region_marker : RegionMarkerLike
            Either a `RegionMarker` object or `(label, coordinates)` tuple,
            where the label must be an `int` and the coordinates a 2- or
            3-element numpy array.
        """
        if not isinstance(region_marker, RegionMarker):
            label, coordinates = region_marker
            region_marker = RegionMarker(label, coordinates)

        self.region_markers.append(region_marker)

    def add_region_markers(self, region_markers: Sequence[RegionMarkerLike]):
        """Add marker to list of region markers.

        Parameters
        ----------
        region_markers : List[RegionMarkerLike]
            List of region markers passed to `.add_region_marker`.
        """
        for region_marker in region_markers:
            self.add_region_marker(region_marker)

    def to_meshio(self) -> 'meshio.Mesh':
        """Return instance of `meshio.Mesh`."""
        cells = [
            (self._cell_type, self.cells),
        ]

        mesh = meshio.Mesh(self.points, cells)

        for key, value in self.cell_data.items():
            mesh.cell_data[key] = [value]

        return mesh

    @classmethod
    def from_meshio(cls, mesh: 'meshio.Mesh'):
        """Return `BaseMesh` from meshio object."""
        points = mesh.points
        cells = mesh.cells[0].data
        cell_data = {}

        for key, value in mesh.cell_data.items():
            # PyVista chokes on ':ref' in cell_data
            key = key.replace(':ref', 'Ref')
            cell_data[key] = value[0]

        return BaseMesh.create(points=points, cells=cells, **cell_data)

    @classmethod
    def create(cls, points, cells, **cell_data):
        """Class dispatcher."""
        cell_dimensions = cells.shape[1]
        if cell_dimensions == 2:
            item_class = LineMesh
        elif cell_dimensions == 3:
            item_class = TriangleMesh
        elif cell_dimensions == 4:
            item_class = TetraMesh
        else:
            item_class = cls
        return item_class(points=points, cells=cells, **cell_data)

    def write(self, *args, **kwargs):
        """Simple wrapper around `meshio.write`."""
        self.to_meshio().write(*args, **kwargs)

    @classmethod
    def read(cls, filename, **kwargs):
        """Simple wrapper around `meshio.read`."""
        mesh = meshio.read(filename, **kwargs)
        return cls.from_meshio(mesh)

    def to_pyvista_unstructured_grid(self) -> 'pv.PolyData':
        """Return instance of `pyvista.UnstructuredGrid`.

        References
        ----------
        https://docs.pyvista.org/core/point-grids.html#pv-unstructured-grid-class-methods
        """
        return pv.from_meshio(self.to_meshio())

    def plot(self, **kwargs):
        raise NotImplementedError

    def plot_mpl(self, **kwargs):
        raise NotImplementedError

    def plot_itk(self, **kwargs):
        """Wrapper for `pyvista.plot_itk`.

        Parameters
        ----------
        **kwargs
            Extra keyword arguments passed to `pyvista.plot_itk`
        """
        pv.plot_itk(self.to_meshio(), **kwargs)

    def plot_pyvista(self, **kwargs):
        """Wrapper for `pyvista.plot`.

        Parameters
        ----------
        **kwargs
            Extra keyword arguments passed to `pyvista.plot`
        """
        pv.plot(self.to_meshio(), **kwargs)

    @property
    def cell_centers(self):
        """Return centers of cells (mean of corner points)."""
        return self.points[self.cells].mean(axis=1)

    def get_cell_data(self, key: str, default_value: float = 0) -> np.ndarray:
        """Get cell data with optional default value.

        Parameters
        ----------
        key : str
            Key of the cell data to retrieve.
        default_value : float, optional
            Optional default value (if cell data does not exist)

        Returns
        -------
        np.ndarray
        """
        try:
            return self.cell_data[self.default_key]
        except KeyError:
            return np.ones(len(self.cells), dtype=int) * default_value

    @property
    def zero_labels(self):
        """Return zero labels as fallback."""
        return np.zeros(len(self.cells), dtype=int)

    @property
    def labels(self):
        """Shortcut for cell labels."""
        try:
            return self.cell_data[self.default_key]
        except KeyError:
            return self.zero_labels

    @labels.setter
    def labels(self, data: np.ndarray):
        """Shortcut for setting cell labels."""
        self.cell_data[self.default_key] = data

    @property
    def dimensions(self):
        """Return number of dimensions for point data."""
        return self.points.shape[1]


class LineMesh(BaseMesh):
    _cell_type = 'line'

    def plot_mpl(self,
                 ax: plt.Axes = None,
                 key: str = None,
                 **kwargs) -> plt.Axes:
        """Simple line mesh plot using `matplotlib`.

        Parameters
        ----------
        ax : plt.Axes, optional
            Axes to use for plotting.
        label : str, optional
            Label of cell data item to plot.
        **kwargs
            Extra keyword arguments passed to `.mpl.lineplot`

        Returns
        -------
        plt.Axes
        """
        from .mpl.lineplot import lineplot

        if not ax:
            fig, ax = plt.subplots()

        if key is None:
            try:
                key = tuple(self.cell_data.keys())[0]
            except IndexError:
                pass

        # https://github.com/python/mypy/issues/9430
        cell_data = self.cell_data.get(key, self.zero_labels)  # type: ignore

        for cell_data_val in np.unique(cell_data):
            vert_x, vert_y = self.points.T

            name = self.number_to_field.get(cell_data_val, cell_data_val)

            lineplot(
                ax,
                x=vert_y,
                y=vert_x,
                lines=self.cells,
                mask=cell_data != cell_data_val,
                label=name,
            )

        if self.region_markers:
            mark_x, mark_y = np.array(
                [m.coordinates for m in self.region_markers]).T
            ax.scatter(mark_y,
                       mark_x,
                       marker='*',
                       color='red',
                       label='Region markers')

        ax.set_title(f'{self._cell_type} mesh')
        ax.axis('equal')

        ax.legend(title=key)

        return ax

    def triangulate(self, opts: str = 'pq30Aa100') -> MeshContainer:
        """Triangulate mesh using `triangle`."""
        from .mesh2d.helpers import simple_triangulate
        points = self.points
        segments = self.cells
        regions = [[*m.coordinates, m.label, 0] for m in self.region_markers]

        return simple_triangulate(points=points,
                                  segments=segments,
                                  regions=regions,
                                  opts=opts)


class TriangleMesh(BaseMesh, PruneZ0Mixin):
    _cell_type = 'triangle'

    def plot(self, **kwargs):
        """Shortcut for `.plot_mpl`."""
        if self.dimensions == 2:
            self.plot_mpl(**kwargs)
        else:
            self.plot_itk(**kwargs)

    def plot_mpl(self,
                 ax: plt.Axes = None,
                 key: str = None,
                 **kwargs) -> plt.Axes:
        """Simple mesh plot using `matplotlib`.

        Parameters
        ----------
        ax : plt.Axes, optional
            Axes to use for plotting.
        key : str, optional
            Label of cell data item to plot, defaults to the
            first key in `.cell_data`.
        fields : dict
            Maps cell data value to string for legend.
        **kwargs
            Extra keyword arguments passed to `ax.triplot`

        Returns
        -------
        plt.Axes
        """
        if not ax:
            fig, ax = plt.subplots()

        if key is None:
            try:
                key = tuple(self.cell_data.keys())[0]
            except IndexError:
                pass

        # https://github.com/python/mypy/issues/9430
        cell_data = self.cell_data.get(key, self.zero_labels)  # type: ignore

        for cell_data_val in np.unique(cell_data):
            vert_x, vert_y = self.points.T

            name = self.number_to_field.get(cell_data_val, cell_data_val)

            ax.triplot(vert_y,
                       vert_x,
                       triangles=self.cells,
                       mask=cell_data != cell_data_val,
                       label=name)

        ax.set_title(f'{self._cell_type} mesh')
        ax.axis('equal')

        _legend_with_triplot_fix(ax, title=key)

        return ax

    def to_trimesh(self):
        """Return instance of `trimesh.Trimesh`."""
        import trimesh
        return trimesh.Trimesh(vertices=self.points, faces=self.cells)

    def to_open3d(self) -> 'open3d.geometry.TriangleMesh':
        """Return instance of `open3d.geometry.TriangleMesh`."""
        import open3d
        return open3d.geometry.TriangleMesh(
            vertices=open3d.utility.Vector3dVector(self.points),
            triangles=open3d.utility.Vector3iVector(self.cells))

    def to_polydata(self) -> 'pv.PolyData':
        """Return instance of `pyvista.Polydata`."""
        points = self.points
        cells = self.cells
        # preprend 3 to indicate number of points per cell
        stacked_cells = np.hstack(np.insert(cells, 0, values=3, axis=1))
        return pv.PolyData(points, stacked_cells, n_faces=len(cells))

    @classmethod
    def from_open3d(cls, mesh: 'open3d.geometry.TriangleMesh') -> TriangleMesh:
        """Return instance of `TriangleMesh` from open3d."""
        points = np.asarray(mesh.vertices)
        cells = np.asarray(mesh.triangles)
        return cls(points=points, cells=cells)

    @classmethod
    def from_scipy(cls, mesh: 'scipy.spatial.qhull.Delaunay') -> TriangleMesh:
        """Return instance of `TriangleMesh` from `scipy.spatial.Delaunay`
        object."""
        points = mesh.points
        cells = mesh.simplices
        return cls(points=points, cells=cells)

    @classmethod
    def from_trimesh(cls, mesh) -> TriangleMesh:
        """Return instance of `TriangleMesh` from trimesh."""
        return cls(points=mesh.vertices, cells=mesh.faces)

    @classmethod
    def from_triangle_dict(cls, dct: dict) -> TriangleMesh:
        """Return instance of `TriangleMesh` from triangle results dict."""
        points = dct['vertices']
        cells = dct['triangles']
        mesh = cls(points=points, cells=cells)

        return mesh

    def optimize(self,
                 *,
                 method='CVT (block-diagonal)',
                 tol: float = 1.0e-3,
                 max_num_steps: int = 10,
                 **kwargs) -> TriangleMesh:
        """Optimize mesh using `optimesh`.

        Parameters
        ----------
        method : str, optional
            Method name
        tol : float, optional
            Tolerance
        max_num_steps : int, optional
            Maximum number of optimization steps.
        **kwargs
            Arguments to pass to `optimesh.optimize_points_cells`

        Returns
        -------
        TriangleMesh
        """
        import optimesh
        points, cells = optimesh.optimize_points_cells(
            X=self.points,
            cells=self.cells,
            method=method,
            tol=tol,
            max_num_steps=max_num_steps,
            **kwargs,
        )
        return TriangleMesh(points=points, cells=cells)

    def tetrahedralize(self, **kwargs) -> 'TetraMesh':
        """Tetrahedralize a contour.

        Parameters
        ----------
        **kwargs
            Keyword arguments passed to `nanomesh.tetgen.tetrahedralize`.

        Returns
        -------
        mesh : TetraMesh
            Tetrahedralized mesh.
        """
        from nanomesh import tetgen
        mesh = tetgen.tetrahedralize(self, **kwargs)
        return mesh


class TetraMesh(BaseMesh):
    _cell_type = 'tetra'

    def to_open3d(self):
        """Return instance of `open3d.geometry.TetraMesh`."""
        import open3d
        return open3d.geometry.TetraMesh(
            vertices=open3d.utility.Vector3dVector(self.points),
            tetras=open3d.utility.Vector4iVector(self.cells))

    @classmethod
    def from_open3d(cls, mesh) -> TetraMesh:
        """Return instance of `TetraMesh` from open3d."""
        points = np.asarray(mesh.vertices)
        cells = np.asarray(mesh.tetras)
        return cls(points=points, cells=cells)

    @classmethod
    def from_pyvista_unstructured_grid(cls, grid: 'pv.UnstructuredGrid'):
        """Return infance of `TetraMesh` from `pyvista.UnstructuredGrid`."""
        assert grid.cells[0] == 4
        cells = grid.cells.reshape(grid.n_cells, 5)[:, 1:]
        points = np.array(grid.points)
        return cls(points=points, cells=cells)

    def plot(self, **kwargs):
        """Shortcut for `.plot_pyvista`."""
        self.plot_pyvista(**kwargs)

    def plot_pyvista(self, **kwargs):
        """Show grid using `pyvista`.

        Parameters
        ----------
        **kwargs
            Keyword arguments passed to `pyvista.Plotter().add_mesh`.
        """
        self.to_pyvista_unstructured_grid().plot(**kwargs)

    def plot_submesh(
        self,
        index: int = None,
        along: str = 'x',
        invert: bool = False,
        show: bool = True,
        backend: str = None,
        **kwargs,
    ):
        """Show submesh using `pyvista`.

        Parameters
        ----------
        index : int, optional
            Index of where to cut the mesh. Shows all tetrahedra
            with cell center < index. Picks the half-way
            point along the axis by default.
        along : str, optional
            Direction along which to cut.
        invert : bool, optional
            Invert the cutting operation, and show all tetrahedra with
            cell center > index.
        show : bool, optional
            If true, show the plot
        **kwargs:
            Keyword arguments passed to `pyvista.Plotter().add_mesh`.

        plotter : `pyvista.Plotter`
            Return plotter instance.
        """
        grid = self.to_pyvista_unstructured_grid()

        # get cell centroids
        cells = grid.cells.reshape(-1, 5)[:, 1:]
        cell_center = grid.points[cells].mean(1)

        # extract cells below index
        axis = 'zyx'.index(along)

        if index is None:
            # pick half-way point
            i, j = axis * 2, axis * 2 + 2
            index = np.mean(grid.bounds[i:j])

        mask = cell_center[:, axis] < index

        if invert:
            mask = ~mask

        cell_ind = mask.nonzero()[0]
        subgrid = grid.extract_cells(cell_ind)

        plotter = pv.Plotter()
        plotter.add_mesh(subgrid, **kwargs)

        if show:
            plotter.show(jupyter_backend=backend)

        return plotter



import logging
import sys

from .mesh import LineMesh, TetraMesh, TriangleMesh
from .mesh2d import Mesher2D, compare_mesh_with_image, simple_triangulate
from .mesh3d import Mesher3D
from .mesh_container import MeshContainer
from .plane import Plane
from .volume import Volume

logging.basicConfig(format='%(message)s',
                    level=logging.INFO,
                    stream=sys.stdout)

__author__ = 'Nicolas Renaud'
__email__ = 'n.renaud@esciencecenter.nl'
__version__ = '0.5.0'

__all__ = [
    '__author__',
    '__email__',
    '__version__',
    'compare_mesh_with_image',
    'LineMesh',
    'Mesher2D',
    'Mesher3D',
    'MeshContainer',
    'Plane',
    'simple_triangulate',
    'TetraMesh',
    'TriangleMesh',
    'Volume',
]



from dataclasses import dataclass
from typing import Callable, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np

from .mesh import BaseMesh


class Metric:
    """Factory function for metrics using `pyvista`.

    Parameters
    ----------
    metric : str
        Metric to calculate. For a full list,
        see `pvmesh.compute_cell_quality`.
    """
    def __init__(self, metric: str):
        super().__init__()
        self.metric = metric

    def __call__(self, mesh: BaseMesh) -> np.ndarray:
        grid = mesh.to_pyvista_unstructured_grid()
        ret = grid.compute_cell_quality(self.metric)
        quality = ret.cell_data['CellQuality']
        return np.array(quality)


# Functions are available, but give undefined values
# aspect_beta = Metric('aspect_beta')
# aspect_gamma = Metric('aspect_gamma')
# collapse_ratio = Metric('collapse_ratio')
# diagonal = Metric('diagonal')
# dimension = Metric('dimension')
# distortion = Metric('distortion')
# jacobian = Metric('jacobian')
# max_aspect_frobenius = Metric('max_aspect_frobenius')
# max_edge_ratio = Metric('max_edge_ratio')
# med_aspect_frobenius = Metric('med_aspect_frobenius')
# oddy = Metric('oddy')
# shear = Metric('shear')
# shear_and_size = Metric('shear_and_size')
# skew = Metric('skew')
# stretch = Metric('stretch')
# taper = Metric('taper')
# volume = Metric('volume')
# warpage = Metric('warpage')

area = Metric('area')
aspect_frobenius = Metric('aspect_frobenius')
aspect_ratio = Metric('aspect_ratio')
condition = Metric('condition')
distortion = Metric('distortion')
max_angle = Metric('max_angle')
min_angle = Metric('min_angle')
radius_ratio = Metric('radius_ratio')
relative_size_squared = Metric('relative_size_squared')
scaled_jacobian = Metric('scaled_jacobian')
shape = Metric('shape')
shape_and_size = Metric('shape_and_size')


def max_min_edge_ratio(mesh: BaseMesh) -> np.ndarray:
    """Place holder, updated dynamically."""
    cell_points = mesh.points[mesh.cells]
    diff = cell_points - np.roll(cell_points, shift=1, axis=1)
    lengths = np.linalg.norm(diff, axis=2)
    return lengths.max(axis=1) / lengths.min(axis=1)


@dataclass
class MetricDescriptor:
    name: str
    description: str
    units: str
    func: Callable
    optimal: Optional[Tuple[float, float]] = None
    range: Optional[Tuple[float, float]] = None

    @property
    def label(self):
        """Return label with units."""
        string = self.name
        if units := self.units:
            string += f' ({units})'
        return string


# coreform.com/cubit_help/mesh_generation/mesh_quality_assessment/triangular_metrics.htm
# vtk.org/doc/nightly/html/classvtkMeshQuality.html#aefa3db78933a64e68c2718cf83eac3c5
# www.feflow.info/html/help73/feflow/09_Parameters/Auxiliary_Data/condition_number.html
_metric_dispatch = {
    'area':
    MetricDescriptor(
        name='Triangle area',
        description='Calculate the area of a triangle.',
        units='px^2',
        optimal=None,
        range=(0, np.inf),
        func=area,
    ),
    'aspect_frobenius':
    MetricDescriptor(
        name='Frobenius aspect',
        description=(
            'Calculate the Frobenius condition number of the '
            'transformation matrix from an equilateral triangle to a triangle.'
        ),
        units='',
        optimal=None,
        range=None,
        func=aspect_frobenius,
    ),
    'aspect_ratio':
    MetricDescriptor(
        name='Aspect ratio',
        description='Calculate the aspect ratio of a triangle.',
        units='',
        optimal=None,
        range=None,
        func=aspect_ratio,
    ),
    'condition':
    MetricDescriptor(
        name='Condition number',
        description='Calculate the condition number of a triangle.',
        units='',
        optimal=(1, 1.3),
        range=(1, np.inf),
        func=condition,
    ),
    'max_angle':
    MetricDescriptor(
        name='Maximum angle',
        description=(
            'Calculate the maximal (nonoriented) angle of a triangle.'),
        units='degrees',
        optimal=(60, 90),
        range=(60, 180),
        func=max_angle,
    ),
    'min_angle':
    MetricDescriptor(
        name='Minimum angle',
        description=(
            'Calculate the minimal (nonoriented) angle of a triangle.'),
        units='degrees',
        optimal=(30, 60),
        range=(0, 60),
        func=min_angle,
    ),
    'radius_ratio':
    MetricDescriptor(
        name='Radius ratio',
        description=(
            'Calculate the radius ratio of a triangle. The radius ratio of a '
            'triangle $t$ is: $\frac{R}{2r}$, where $R$ and $r$ respectively '
            'denote the circumradius and the inradius of $t$.'),
        units='',
        optimal=(1.0, 2.0),
        range=(1, np.inf),
        func=radius_ratio,
    ),
    'scaled_jacobian':
    MetricDescriptor(
        name='Scaled Jacobian',
        description='Calculate the scaled Jacobian of a triangle.',
        units='',
        optimal=(0.2, 1.0),
        range=(-1, 1),
        func=scaled_jacobian,
    ),
    'shape':
    MetricDescriptor(
        name='Shape',
        description='Calculate the shape of a triangle.',
        units='',
        optimal=(0.25, 1.0),
        range=(0, 1),
        func=shape,
    ),
    'relative_size_squared':
    MetricDescriptor(
        name='Relative size',
        description='Calculate the square of the relative size of a triangle.',
        units='',
        optimal=(0.25, 1.0),
        range=(0, 1),
        func=relative_size_squared,
    ),
    'shape_and_size':
    MetricDescriptor(
        name='Shape and size',
        description=(
            'Calculate the product of shape and relative size of a triangle.'),
        units='',
        optimal=(0.25, 1.0),
        range=(0, 1),
        func=shape_and_size,
    ),
    'distortion':
    MetricDescriptor(
        name='Distortion',
        description='Calculate the distortion of a triangle.',
        units='px^2',
        optimal=(0.6, 1.0),
        range=(0, 1),
        func=distortion,
    ),
    'max_min_edge_ratio':
    MetricDescriptor(
        name='Ratio max/min edge',
        description=('Calculate the ratio between the longest '
                     'and shortest edge lengths of a triangle.'),
        units='',
        optimal=(1.0, 2.0),
        range=(1, np.inf),
        func=max_min_edge_ratio,
    ),
}

# patch docstrings
for descriptor in _metric_dispatch.values():
    descriptor.func.__doc__ = f"""{descriptor.description}

Parameters
----------
mesh : BaseMesh
    Input mesh

Returns
-------
quality : np.ndarray
    Array with cell qualities.
    """


def calculate_all_metrics(mesh: BaseMesh, inplace: bool = False) -> dict:
    """Calculate all available metrics.

    Parameters
    ----------
    mesh : BaseMesh
        Input mesh
    inplace : bool, optional
        Updates the `cell_data` attribute on the mesh with the metrics.

    Returns
    -------
    metrics : dict
        Return a dict with all the metrics
    """
    metrics = {}
    for metric, descriptor in _metric_dispatch.items():
        quality = descriptor.func(mesh)  # type: ignore
        metrics[metric] = quality

        if inplace:
            mesh.cell_data[metric] = quality

    return metrics


def histogram(
    mesh: BaseMesh,
    *,
    metric: str,
    ax: plt.Axes = None,
    **kwargs,
) -> plt.Axes:
    """Create a mesh plot with the cells are colored by the cell quality.

    Parameters
    ----------
    mesh : BaseMesh
        Input mesh
    metric : str
        Metric to calculate.
    ax : `matplotlib.Axes`
        If specified, `ax` will be used to create the subplot.
    vmin, vmax : int, float
        Set the lower/upper boundary for the color value.
    cmap : str
        Set the color map.
    **kwargs
        Keyword arguments passed on to `ax.hist`.

    Returns
    -------
    ax : `matplotlib.Axes`
    """
    kwargs.setdefault('bins', 50)
    kwargs.setdefault('rwidth', 0.8)

    descriptor = _metric_dispatch[metric]
    quality = descriptor.func(mesh)  # type: ignore

    fig, ax = plt.subplots()
    n, bins, patches = ax.hist(quality, **kwargs)
    ax.set_title(f'Histogram of {descriptor.name.lower()}')

    if optimal_range := descriptor.optimal:
        ax.axvspan(*optimal_range, color='limegreen', zorder=0, alpha=0.1)

    xlabel = descriptor.label
    ax.set_xlabel(xlabel)

    ylabel = 'probability' if kwargs.get('density') else 'frequency'
    ax.set_ylabel(ylabel)

    return ax


def plot2d(
    mesh: BaseMesh,
    *,
    metric: str,
    ax: plt.Axes = None,
    **kwargs,
) -> plt.Axes:
    """Create a mesh plot with the cells are colored by the cell quality.

    Parameters
    ----------
    mesh : BaseMesh
        Input mesh
    metric : str
        Metric to calculate.
    ax : `matplotlib.Axes`
        If specified, `ax` will be used to create the subplot.
    vmin, vmax : int, float
        Set the lower/upper boundary for the color value.
        Defaults to the 1st and 99th percentile, respectively.
    cmap : str
        Set the color map.
    **kwargs
        Keyword arguments passed on to `ax.tripcolor`.

    Returns
    -------
    ax : `matplotlib.Axes`
    """
    descriptor = _metric_dispatch[metric]
    quality = descriptor.func(mesh)  # type: ignore

    kwargs.setdefault('vmin', np.percentile(quality, 1))
    kwargs.setdefault('vmax', np.percentile(quality, 99))

    fig, ax = plt.subplots()

    x = mesh.points[:, 0]
    y = mesh.points[:, 1]

    cells = mesh.cells

    tpc = ax.tripcolor(x, y, quality, triangles=cells, **kwargs)
    ax.figure.colorbar(tpc)
    ax.axis('scaled')

    ax.set_title(f'Triplot of {descriptor.name.lower()}')

    return ax



import logging
from abc import ABC, abstractmethod
from typing import Union

import numpy as np

from .plane import Plane
from .volume import Volume

logger = logging.getLogger(__name__)


class BaseMesher(ABC):
    def __init__(self, image: Union[np.ndarray, Plane, Volume]):
        if isinstance(image, (Plane, Volume)):
            image = image.image

        self.image_orig = image
        self.image = image

    @abstractmethod
    def show_contour(self):
        ...

    @abstractmethod
    def generate_contour(self):
        ...



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from nanomesh.utils import pairwise

from ..region_markers import RegionMarker
from .bounding_box import BoundingBox

if TYPE_CHECKING:
    from nanomesh.mesh import TriangleMesh


def pad(mesh: TriangleMesh,
        *,
        side: str,
        width: int,
        label: int = None) -> TriangleMesh:
    """Pad a triangle mesh (3D).

    Parameters
    ----------
    mesh : TriangleMesh
        The mesh to pad.
    side : str
        Side to pad, must be one of `left`, `right`, `top`,
        `bottom`, `back`, `front`.
    width : int
        Width of the padded area.
    label : int, optional
        The label to assign to the padded area. If not defined, generates the
        next unique label based on the existing ones.

    Returns
    -------
    new_mesh : TriangleMesh
        Padded tetrahedral mesh.

    Raises
    ------
    ValueError
        When the value of `side` is invalid.
    """
    if label is None:
        label = mesh.labels.max() + 1

    if width == 0:
        return mesh

    bbox = BoundingBox.from_points(mesh.points)

    if side == 'top':
        edge_col = 2
        edge_value = bbox.zmax
        extra_coords = np.array([
            [bbox.xmin, bbox.ymin, bbox.zmax + width],
            [bbox.xmin, bbox.ymax, bbox.zmax + width],
            [bbox.xmax, bbox.ymin, bbox.zmax + width],
            [bbox.xmax, bbox.ymax, bbox.zmax + width],
        ])
        column_order = (0, 1, 1, 0)
    elif side == 'bottom':
        edge_col = 2
        edge_value = bbox.zmin
        extra_coords = np.array([
            [bbox.xmin, bbox.ymin, bbox.zmin - width],
            [bbox.xmin, bbox.ymax, bbox.zmin - width],
            [bbox.xmax, bbox.ymin, bbox.zmin - width],
            [bbox.xmax, bbox.ymax, bbox.zmin - width],
        ])
        column_order = (0, 1, 1, 0)
    elif side == 'left':
        edge_col = 1
        edge_value = bbox.ymin
        extra_coords = np.array([
            [bbox.xmin, bbox.ymin - width, bbox.zmin],
            [bbox.xmin, bbox.ymin - width, bbox.zmax],
            [bbox.xmax, bbox.ymin - width, bbox.zmin],
            [bbox.xmax, bbox.ymin - width, bbox.zmax],
        ])
        column_order = (0, 2, 2, 0)
    elif side == 'right':
        edge_col = 1
        edge_value = bbox.ymax
        extra_coords = np.array([
            [bbox.xmin, bbox.ymax + width, bbox.zmin],
            [bbox.xmin, bbox.ymax + width, bbox.zmax],
            [bbox.xmax, bbox.ymax + width, bbox.zmin],
            [bbox.xmax, bbox.ymax + width, bbox.zmax],
        ])
        column_order = (0, 2, 2, 0)
    elif side == 'front':
        edge_col = 0
        edge_value = bbox.xmin
        extra_coords = np.array([
            [bbox.xmin - width, bbox.ymin, bbox.zmin],
            [bbox.xmin - width, bbox.ymin, bbox.zmax],
            [bbox.xmin - width, bbox.ymax, bbox.zmin],
            [bbox.xmin - width, bbox.ymax, bbox.zmax],
        ])
        column_order = (1, 2, 2, 1)
    elif side == 'back':
        edge_col = 0
        edge_value = bbox.xmax
        extra_coords = np.array([
            [bbox.xmax + width, bbox.ymin, bbox.zmin],
            [bbox.xmax + width, bbox.ymin, bbox.zmax],
            [bbox.xmax + width, bbox.ymax, bbox.zmin],
            [bbox.xmax + width, bbox.ymax, bbox.zmax],
        ])
        column_order = (1, 2, 2, 1)
    else:
        raise ValueError('Side must be one of `right`, `left`, `bottom`'
                         f'`top`, `front`, `back`. Got {side=}')

    n_points = len(mesh.points)
    points = np.vstack([mesh.points, extra_coords])

    new_triangles = [
        np.array((0, 1, 2)) + n_points,
        np.array((3, 1, 2)) + n_points,
    ]

    for corner, col in zip(extra_coords, column_order):
        connect_to = np.argwhere((points[:, edge_col] == edge_value)
                                 & (points[:, col] == corner[col]))

        additional_points = np.argwhere(
            extra_coords[:, col] == corner[col]) + n_points

        first, last = additional_points
        first_point = points[first]

        sorted_by_distance = np.argsort(
            np.linalg.norm(first_point - points[connect_to].squeeze(), axis=1))
        connect_to = connect_to[sorted_by_distance]
        connect_to = np.vstack([connect_to, last]).squeeze()

        for pair in pairwise(connect_to):
            tri = np.hstack((first, pair))
            new_triangles.append(tri)

    new_triangles = np.array(new_triangles).squeeze()

    cells = np.vstack([mesh.cells, new_triangles])

    new_mesh = mesh.__class__(
        points=points,
        cells=cells,
        region_markers=mesh.region_markers,
    )

    # add marker for new region
    center = extra_coords.mean(axis=0)
    center[col] = (center[col] + edge_value) / 2

    new_mesh.add_region_marker(RegionMarker(label, center))

    return new_mesh



import matplotlib.pyplot as plt
import numpy as np


def lineplot(ax: plt.Axes,
             *,
             x: np.ndarray,
             y: np.ndarray,
             lines: np.ndarray,
             mask: np.ndarray = None,
             label: str = None,
             **kwargs):
    """Plot collection of lines, similar to `ax.triplot`.

    Parameters
    ----------
    ax : plt.Axes
        Description
    x : (n, 1) np.ndarray
        x-coordinates of points.
    y : (n, 1) np.ndarray
        y-coordinates of points.
    lines : (m, 2) np.ndarray
        Integer array describing the connected lines.
    mask : (m, 1) np.ndarray, optional
        Mask for line segments.
    label : str, optional
        Label for legend.
    **kwargs : dict
        Extra keywords arguments passed to `ax.plot`

    Returns
    -------
    list of `matplotlib.lines.Line2D`
        A list of lines representing the plotted data.
    """
    kwargs.setdefault('marker', '.')

    if mask is not None:
        lines = lines[~mask.squeeze()]

    lines_x = np.insert(x[lines], 2, np.nan, axis=1)
    lines_y = np.insert(y[lines], 2, np.nan, axis=1)
    return ax.plot(lines_x.ravel(), lines_y.ravel(), label=label, **kwargs)



import numpy as np
from matplotlib.widgets import PolygonSelector


class PolygonSelectorWithSnapping(PolygonSelector):
    """Select a polygon region of an axes with snapping to points.

    For usage details see `matplotlib.widgets.PolygonSelector`

    Parameters
    ----------
    snap_to : (n,2) np.ndarray
        List of points to snap to .
    *args : list
        Arguments passed to `matplotlib.widgets.PolygonSelector`.
    **kwargs : dict
        Keyword arguments passed to `matplotlib.widgets.PolygonSelector`.
        The parent axes for the widget.
    """
    def __init__(
        self,
        *args,
        snap_to: np.ndarray = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.snapped = False
        self._snap_to_points = snap_to
        self.SNAPPING = (snap_to is not None)

    def _release(self, event):
        """Button release event handler."""
        # Release active tool handle.
        if self._active_handle_idx >= 0:
            self._active_handle_idx = -1

        # Complete the polygon.
        elif (len(self._xs) > 3 and self._xs[-1] == self._xs[0]
              and self._ys[-1] == self._ys[0]):
            self._selection_completed = True

        elif self.snapped is not None:
            self._xs.insert(-1, self._xs[-1])
            self._ys.insert(-1, self._ys[-1])

        # Place new vertex.
        elif (not self._selection_completed and 'move_all' not in self.state
              and 'move_vertex' not in self.state):
            self._xs.insert(-1, event.xdata)
            self._ys.insert(-1, event.ydata)

        if self._selection_completed:
            self.onselect(self.verts)

    def _onmove(self, event):
        """Cursor move event handler."""
        # Move the active vertex (ToolHandle).
        if self._active_handle_idx >= 0:
            idx = self._active_handle_idx
            self._xs[idx], self._ys[idx] = event.xdata, event.ydata
            # Also update the end of the polygon line if the first vertex is
            # the active handle and the polygon is completed.
            if idx == 0 and self._selection_completed:
                self._xs[-1], self._ys[-1] = event.xdata, event.ydata

        # Move all vertices.
        elif 'move_all' in self.state and self.eventpress:
            dx = event.xdata - self.eventpress.xdata
            dy = event.ydata - self.eventpress.ydata
            for k in range(len(self._xs)):
                self._xs[k] = self._xs_at_press[k] + dx
                self._ys[k] = self._ys_at_press[k] + dy

        # Do nothing if completed or waiting for a move.
        elif (self._selection_completed or 'move_vertex' in self.state
              or 'move_all' in self.state):
            return

        # Position pending vertex.
        else:
            # Calculate distance to the start vertex.
            x0, y0 = self.line.get_transform().transform(
                (self._xs[0], self._ys[0]))
            v0_dist = np.hypot(x0 - event.x, y0 - event.y)

            if self.SNAPPING:
                # Calculate nearest point to snap to.
                self._tf_snap_to = self.line.get_transform().transform(
                    self._snap_to_points)
                vn_dists = np.hypot(self._tf_snap_to[:, 0] - event.x,
                                    self._tf_snap_to[:, 1] - event.y)
                vn_ind = np.argmin(vn_dists)

            # Lock on to the start vertex if near it and ready to complete.
            if len(self._xs) > 3 and v0_dist < self.vertex_select_radius:
                self._xs[-1], self._ys[-1] = self._xs[0], self._ys[0]

            # Lock on to pre-defined point if near it
            elif self.SNAPPING and (vn_dists[vn_ind] <
                                    self.vertex_select_radius):
                self._xs[-1], self._ys[-1] = self._snap_to_points[vn_ind]
                self._snapped = True

            else:
                self._xs[-1], self._ys[-1] = event.xdata, event.ydata
                self._snapped = False

        self._draw_polygon()



import logging
import os
from pathlib import Path
from typing import Callable, Tuple, Union

import meshio
import numpy as np

from .base_image import BaseImage
from .io import load_vol
from .plane import Plane

logger = logging.getLogger(__name__)


class Volume(BaseImage):
    @classmethod
    def load(cls, filename: os.PathLike, **kwargs) -> 'Volume':
        """Load the data. Supported filetypes: `.npy`, `.vol`.

        For memory mapping, use `mmap_mode='r'`. Memory-mapped
            files are used for accessing small segments of large files on
            disk, without reading the entire file into memory. Note that this
            can still result in some slow / unexpected behaviour with some
            operations.

            More info:
            https://numpy.org/doc/stable/reference/generated/numpy.memmap.html

        Parameters
        ----------
        filename : PathLike
            Name of the file to load.
        **kwargs : dict
            Extra keyword arguments passed onto data readers.

        Returns
        -------
        Volume
            Instance of this class.

        Raises
        ------
        IOError
            Raised if the file extension is unknown.
        """
        filename = Path(filename)
        suffix = filename.suffix.lower()

        if suffix == '.npy':
            array = np.load(filename, **kwargs)
        elif suffix == '.vol':
            array = load_vol(filename, **kwargs)
        else:
            raise IOError(f'Unknown file extension: {suffix}')
        return cls(array)

    def apply(self, function: Callable, **kwargs) -> 'Volume':
        """Apply function to `.image` array. Return an instance of `Volume` if
        the result is a 3D image, otherwise return the result of the operation.

        Parameters
        ----------
        function : callable
            Function to apply to `self.image`.
        **kwargs
            Keyword arguments to pass to `function`.

        Returns
        -------
        Volume
            New instance of `Volume`.
        """
        return super().apply(function, **kwargs)

    def show_slice(self, **kwargs):
        """Show slice using `nanomesh.utils.SliceViewer`.

        Extra arguments are passed on.
        """
        from nanomesh.utils import SliceViewer
        sv = SliceViewer(self.image, **kwargs)
        sv.interact()
        return sv

    def show_volume(self, renderer: str = 'itkwidgets', **kwargs) -> None:
        """Show volume using `itkwidgets` or `ipyvolume`.

        Extra keyword arguments (`kwargs`) are passed to
        `itkwidgets.view` or `ipyvolume.quickvolshow`.
        """
        if renderer in ('ipyvolume', 'ipv'):
            import ipyvolume as ipv
            return ipv.quickvolshow(self.image, **kwargs)
        elif renderer in ('itkwidgets', 'itk', 'itkw'):
            import itkwidgets as itkw
            return itkw.view(self.image)
        else:
            raise ValueError(f'No such renderer: {renderer!r}')

    def generate_mesh(self, **kwargs) -> 'meshio.Mesh':
        """Generate mesh from binary (segmented) image.

        Parameters
        ----------
        **kwargs:
            Keyword arguments are passed to `mesh3d.generate_3d_mesh`

        Returns
        -------
        meshio.Mesh
            Description of the mesh.
        """
        from nanomesh.mesh3d import generate_3d_mesh
        return generate_3d_mesh(image=self.image, **kwargs)

    def select_plane(self,
                     x: int = None,
                     y: int = None,
                     z: int = None) -> 'Plane':
        """Select a slice in the volume. Either `x`, `y` or `z` must be
        specified.

        Parameters
        ----------
        x : int, optional
            Index along the x-axis
        y : int, optional
            Index along the y-axis
        z : int, optional
            Index along the z-axis

        Returns
        -------
        Plane
            Return 2D plane representation.

        Raises
        ------
        ValueError
            If none of the `x`, `y`, or `z` arguments have been specified
        """
        index: Tuple[Union[int, slice], ...]

        if x is not None:
            index = np.s_[:, :, x]
        elif y is not None:
            index = np.s_[:, y, :]
        elif z is not None:
            index = np.s_[z, :, :]
        else:
            raise ValueError(
                'One of the arguments `x`, `y`, or `z` must be specified.')

        array = self.image[index]
        return Plane(array)

    def select_subvolume(self,
                         *,
                         xs: tuple = None,
                         ys: tuple = None,
                         zs: tuple = None) -> 'Volume':
        """Select a subvolume from the current volume.

        Each range must include a start and stop value, for example:

        `vol.select_subvolume(xs=(10, 20))` is equivalent to:
        `vol.image[[:,:,10:20]`

        or

        `vol.select_subvolume(ys=(20, 25), zs=(40, 50))` is equivalent to:
        `vol.image[[40:50,20:25,:]`

        Parameters
        ----------
        xs : tuple, optional
            Range to select along y-axis
        ys : tuple, optional
            Range to select along y-axis
        zs : tuple, optional
            Range to select along z-axis

        Returns
        -------
        Plane
            Return 3D volume representation.
        """
        default = slice(None)
        slices = tuple(slice(*r) if r else default for r in (zs, ys, xs))

        array = self.image[slices]
        return Volume(array)



from __future__ import annotations

from collections import defaultdict
from enum import Enum
from types import MappingProxyType
from typing import Dict, List

import meshio
import numpy as np

from .mesh import BaseMesh, PruneZ0Mixin


class CellType(Enum):
    NULL = 0
    LINE = 1
    TRIANGLE = 2
    TETRA = 3


class MeshContainer(meshio.Mesh, PruneZ0Mixin):
    @property
    def number_to_field(self):
        """Mapping from numbers to fields, proxy to `.field_data`."""
        number_to_field = defaultdict(dict)

        for field, (number, dimension) in self.field_data.items():
            dim_name = CellType(dimension).name.lower()
            number_to_field[dim_name][number] = field

        return MappingProxyType(
            {k: MappingProxyType(v)
             for k, v in number_to_field.items()})

    @property
    def field_to_number(self):
        """Mapping from fields to numbers, proxy to `.field_data`."""
        field_to_number = defaultdict(dict)

        for field, (number, dimension) in self.field_data.items():
            dim_name = CellType(dimension).name.lower()
            field_to_number[dim_name][field] = number

        return MappingProxyType(
            {k: MappingProxyType(v)
             for k, v in field_to_number.items()})

    def set_field_data(self, cell_type: str, field_data: Dict[int, str]):
        """Update the values in `.field_data`.

        Parameters
        ----------
        cell_type : str
            Cell type to update the values for.
        field_data : dict
            Dictionary with key-to-number mapping, i.e.
            `field_data={0: 'green', 1: 'blue', 2: 'red'}`
            maps `0` to `green`, etc.
        """
        try:
            input_field_data = dict(self.number_to_field[cell_type])
        except KeyError:
            input_field_data = {}

        input_field_data.update(field_data)

        new_field_data = self.field_data.copy()

        remove_me = []

        for field, (value, field_cell_type) in new_field_data.items():
            if CellType(field_cell_type) == CellType[cell_type.upper()]:
                remove_me.append(field)

        for field in remove_me:
            new_field_data.pop(field)

        for value, field in input_field_data.items():
            CELL_TYPE = CellType[cell_type.upper()].value
            new_field_data[field] = [value, CELL_TYPE]

        self.field_data: Dict[str, List[int]] = new_field_data

    @property
    def cell_types(self):
        """Return cell types in order."""
        return tuple(cell.type for cell in self.cells)

    def set_cell_data(self, cell_type: str, key: str, value):
        """Set `key` to `value` for `cell_type` in `.cell_data_dict`."""
        index = self.cell_types.index(cell_type)
        assert len(value) == len(self.cells_dict[cell_type])

        try:
            self.cell_data[key][index] = value
        except KeyError:
            new_cell_data = []

            # set missing cells to 0
            for i, _ in enumerate(self.cell_types):
                if i == index:
                    new_cell_data.append(value)
                else:
                    new_cell_data.append(
                        np.zeros(len(self.cells[0].data), dtype=int))

            self.cell_data[key] = new_cell_data

    def get_default_type(self) -> str:
        """Try to return highest dimension type.

        Default to first type `cells_dict`.

        Returns
        -------
        cell_type : str
        """
        for type_ in ('tetra', 'triangle', 'line'):
            if type_ in self.cells_dict:
                return type_

        return list(self.cells_dict.keys())[0]

    def get(self, cell_type: str = None) -> BaseMesh:
        """Extract mesh with points/cells of `cell_type`.

        Parameters
        ----------
        cell_type : str, optional
            Element type, such as line, triangle, tetra, etc.

        Returns
        -------
        BaseMesh
            Dataclass with `points`/`cells` attributes
        """
        if not cell_type:
            cell_type = self.get_default_type()

        try:
            cells = self.cells_dict[cell_type]
        except KeyError as e:
            msg = (f'No such cell type: {cell_type!r}. '
                   f'Must be one of {tuple(self.cells_dict.keys())!r}')
            raise KeyError(msg) from e

        points = self.points

        cell_data = self.get_all_cell_data(cell_type)

        fields = self.field_to_number.get(cell_type, None)

        return BaseMesh.create(cells=cells,
                               points=points,
                               fields=fields,
                               **cell_data)

    def get_all_cell_data(self, cell_type: str = None) -> dict:
        """Get all cell data for given `cell_type`.

        Parameters
        ----------
        cell_type : str, optional
            Element type, such as line, triangle, tetra, etc.

        Returns
        -------
        data_dict : dict
            Dictionary with cell data
        """
        if not cell_type:
            cell_type = self.get_default_type()

        data_dict = {}
        for key in self.cell_data:
            new_key = key.replace(':', '-')
            data_dict[new_key] = self.get_cell_data(key, cell_type)

        return data_dict

    def plot(self, cell_type: str = None, **kwargs):
        """Plot data.

        Parameters
        ----------
        cell_type : str, optional
            Cell type to plot.
        **kwargs
            Extra keyword arguments passed to plotting method.
        """
        mesh = self.get(cell_type)
        mesh.plot(**kwargs)

    def plot_mpl(self, cell_type: str = None, **kwargs):
        """Plot data using matplotlib.

        Parameters
        ----------
        cell_type : str, optional
            Cell type to plot.
        **kwargs
            Extra keyword arguments passed to plotting method.
        """
        mesh = self.get(cell_type)
        fields = self.number_to_field.get(mesh._cell_type, None)
        mesh.plot_mpl(fields=fields, **kwargs)

    def plot_itk(self, cell_type: str = None, **kwargs):
        """Plot data using itk.

        Parameters
        ----------
        cell_type : str, optional
            Cell type to plot.
        **kwargs
            Extra keyword arguments passed to plotting method.
        """
        mesh = self.get(cell_type)
        mesh.plot_itk(**kwargs)

    def plot_pyvista(self, cell_type: str = None, **kwargs):
        """Plot data using pyvista.

        Parameters
        ----------
        cell_type : str, optional
            Cell type to plot.
        **kwargs
            Extra keyword arguments passed to plotting method.
        """
        mesh = self.get(cell_type)
        mesh.plot_pyvista(**kwargs)

    @classmethod
    def from_mesh(cls, mesh: BaseMesh):
        """Convert from `BaseMesh` to `MeshContainer`.

        Parameters
        ----------
        mesh : BaseMesh
            Input mesh, must be a subclass of `BaseMesh`.

        Returns
        -------
        MeshContainer
        """
        meshio_mesh = mesh.to_meshio()
        return cls(points=meshio_mesh.points,
                   cells=meshio_mesh.cells,
                   cell_data=meshio_mesh.cell_data)

    @classmethod
    def from_triangle_dict(cls, triangle_dict: dict):
        """Return instance of `MeshContainer` from triangle results dict.

        Parameters
        ----------
        triangle_dict : dict
            Triangle triangulate output dictionary.

        Returns
        -------
        mesh : MeshContainer
        """
        points = triangle_dict['vertices']

        cell_data = {}
        cells = {}

        cells['triangle'] = triangle_dict['triangles']

        point_data = {}
        try:
            point_data['physical'] = triangle_dict['vertex_markers'].squeeze()
        except KeyError:
            pass

        try:
            cells['line'] = triangle_dict['edges']
            # Order must match order of cell_data
            cell_data['physical'] = [
                triangle_dict['triangle_attributes'].squeeze(),
                triangle_dict['edge_markers'].squeeze(),
            ]
        except KeyError:
            pass

        mesh = cls(
            points=points,
            cells=cells,
            cell_data=cell_data,
            point_data=point_data,
        )

        mesh.triangle_dict = triangle_dict

        return mesh

    @classmethod
    def read(cls, *args, **kwargs):
        """Wrapper for `meshio.read`.

        For gmsh:
        - remaps `gmsh:physical` -> `physical`
        - remaps `gmsh:geometrical` -> `geometrical`
        """
        from meshio import read
        mesh = read(*args, **kwargs)

        cell_data = {}
        for key, value in mesh.cell_data.items():
            if key in ('gmsh:physical', 'gmsh:geometrical'):
                key = key.replace('gmsh:', '')

            cell_data[key] = value

        point_data = {}
        for key, value in mesh.point_data.items():
            if key in ('gmsh:physical', 'gmsh:geometrical'):
                key = key.replace('gmsh:', '')

            point_data[key] = value

        ret = cls(
            mesh.points,
            mesh.cells,
            point_data=point_data,
            cell_data=cell_data,
            field_data=mesh.field_data,
            point_sets=mesh.point_sets,
            cell_sets=mesh.cell_sets,
            gmsh_periodic=mesh.gmsh_periodic,
            info=mesh.info,
        )
        ret.prune_z_0()
        return ret

    def write(self, filename, file_format: str = None, **kwargs):
        """Thin wrapper of `meshio.write` to avoid altering class.

        For gmsh:
        - remaps `physical` -> `gmsh:physical`
        - remaps `geometrical` -> `gmsh:geometrical`

        Parameters
        ----------
        filename : str
            File to write to.
        file_format : str, optional
            Specify file format. By default, this is guessed from the
            extension.
        **kwargs
            Extra keyword arguments passed to `meshio.write`.
        """
        from pathlib import Path

        from meshio import write
        from meshio._helpers import _filetype_from_path

        if file_format is None:
            file_format = _filetype_from_path(Path(filename))

        if file_format.startswith('gmsh'):
            cell_data = {}
            for key, value in self.cell_data.items():
                if key in ('physical', 'geometrical'):
                    key = f'gmsh:{key}'

                cell_data[key] = value

            point_data = {}
            for key, value in self.point_data.items():
                if key in ('physical', 'geometrical'):
                    key = f'gmsh:{key}'

                point_data[key] = value
        else:
            cell_data = self.cell_data
            point_data = self.point_data

        out_mesh = meshio.Mesh(
            self.points,
            self.cells,
            point_data=point_data,
            cell_data=cell_data,
            field_data=self.field_data,
            point_sets=self.point_sets,
            cell_sets=self.cell_sets,
            gmsh_periodic=self.gmsh_periodic,
            info=self.info,
        )

        write(filename, mesh=out_mesh, file_format=file_format, **kwargs)



from types import SimpleNamespace

import numpy as np
from matplotlib.patches import Polygon
from scipy.spatial import ConvexHull
from skimage import transform

from .mpl.widgets import PolygonSelectorWithSnapping


def minimum_bounding_rectangle(coords: np.ndarray) -> np.ndarray:
    """Find the smallest bounding rectangle for a set of coordinates.

    Based on: https://stackoverflow.com/a/33619018

    Parameters
    ----------
    coords : (n,2) np.ndarray
        List of coordinates.

    Returns
    -------
    bbox_coords: (4,2) np.ndarray
        List of coordinates representing the corners of the bounding box.
    """
    # get the convex hull for the coords
    hull_coords = coords[ConvexHull(coords).vertices]

    # calculate edge angles
    edges = np.zeros((len(hull_coords) - 1, 2))
    edges = hull_coords[1:] - hull_coords[:-1]

    angles = np.zeros((len(edges)))
    angles = np.arctan2(edges[:, 1], edges[:, 0])

    angles = np.abs(np.mod(angles, np.pi / 2))
    angles = np.unique(angles)

    # find rotation matrices
    rotations = np.vstack([
        np.cos(angles),
        np.cos(angles - np.pi / 2),
        np.cos(angles + np.pi / 2),
        np.cos(angles)
    ]).T
    rotations = rotations.reshape((-1, 2, 2))

    # apply rotations to the hull
    rot_coords = np.dot(rotations, hull_coords.T)

    # find the bounding coords
    min_x = np.nanmin(rot_coords[:, 0], axis=1)
    max_x = np.nanmax(rot_coords[:, 0], axis=1)
    min_y = np.nanmin(rot_coords[:, 1], axis=1)
    max_y = np.nanmax(rot_coords[:, 1], axis=1)

    # find the box with the best area
    areas = (max_x - min_x) * (max_y - min_y)
    best_idx = np.argmin(areas)

    # return the best box
    x1 = max_x[best_idx]
    x2 = min_x[best_idx]
    y1 = max_y[best_idx]
    y2 = min_y[best_idx]
    rotation = rotations[best_idx]

    bbox_coords = np.zeros((4, 2))
    bbox_coords[0] = np.dot([x1, y2], rotation)
    bbox_coords[1] = np.dot([x2, y2], rotation)
    bbox_coords[2] = np.dot([x2, y1], rotation)
    bbox_coords[3] = np.dot([x1, y1], rotation)

    return bbox_coords


def extract_rectangle(image: np.ndarray, *, bbox: np.ndarray):
    """Extract rectangle from image. The image is straightened using an
    Euclidean transform.

    Parameters
    ----------
    image : 2D np.ndarray
        Image to extract rectangle from.
    bbox : (4,2) np.ndarray
        Four coordinate describing the corners of the bounding box.

    Returns
    -------
    warped : 2D np.ndarray
        The warped input image extracted from the bounding box.
    """
    a = int(np.linalg.norm(bbox[0] - bbox[1]))
    b = int(np.linalg.norm(bbox[1] - bbox[2]))

    src = np.array([[0, 0], [0, a], [b, a], [b, 0]])
    dst = np.array(bbox)

    tform3 = transform.EuclideanTransform()
    tform3.estimate(src, dst)
    warped = transform.warp(image, tform3, output_shape=(a, b))

    return warped


class ROISelector:
    ROTATE = True
    """Select a region of interest points in the figure by enclosing them
    within a polygon. A rectangle is fitted to the polygon.

    - Press the 'esc' key to start a new polygon.
    - Hold the 'shift' key to move all of the vertices.
    - Hold the 'ctrl' key to move a single vertex.

    Attributes
    ----------
    bbox : (4,2) np.ndarray
        Coordinates describing the corners of the polygon
    """
    def __init__(self, ax, snap_to: np.ndarray = None):
        self.ax = ax
        self.canvas = ax.figure.canvas
        self.bbox = np.array([[0, 0], [0, 1], [1, 1], [1, 0]])
        self.verts = None

        self.poly = PolygonSelectorWithSnapping(ax,
                                                self.onselect,
                                                snap_to=snap_to)

    def onselect(self, verts):
        """Trigger this function when a polygon is closed."""
        self.verts = np.array(verts)
        self.bbox = self.bounding_rectangle(rotate=self.ROTATE)

        bounds = self.get_bounds()
        self.ax.set_title(f'left {bounds.left:.0f} '
                          f'right {bounds.right:.0f}'
                          f'\ntop {bounds.top:.0f} '
                          f'bottom {bounds.bottom:.0f}')
        self.draw_bbox()
        self.canvas.draw_idle()

    def disconnect(self):
        """Disconnect the selector."""
        self.poly.disconnect_events()
        self.canvas.draw_idle()

    def draw_bbox(self):
        """Draw bounding box as a patch on the image."""
        # remove existing patches in case the roi is modified
        self.ax.patches.clear()
        polygon = Polygon(self.bbox, facecolor='red', alpha=0.3)
        self.ax.add_patch(polygon)

    def get_bounds(self) -> SimpleNamespace:
        """Get bounds of bbox (left, right, top, bottom)."""
        left, top = self.bbox.min(axis=0)
        right, bottom = self.bbox.max(axis=0)
        bounds = SimpleNamespace(left=left,
                                 top=top,
                                 right=right,
                                 bottom=bottom)
        return bounds

    def bounding_rectangle(self, rotate=True) -> np.ndarray:
        """Return bounding rectangle.

        Parameters
        ----------
        rotate : bool, optional
            If True, allow rotation of the bounding box to find the minumum
            bounding rectangle.

        Returns
        -------
        np.ndarray
            Description
        """
        if self.verts is None:
            raise ValueError('No vertices have been selected!')

        if rotate:
            return minimum_bounding_rectangle(self.verts)
        else:
            left, top = self.verts.min(axis=0)
            right, bottom = self.verts.max(axis=0)

            return np.array([[right, bottom], [right, top], [left, top],
                             [left, bottom]])



import time
from itertools import tee

import matplotlib.pyplot as plt
import numpy as np
from ipywidgets import IntSlider, RadioButtons, interact


class SliceViewer:
    """Simple slice viewer for volumes using matplotlib.

    Parameters
    ----------
    data : 3D np.ndarray
        Volume to display.
    update_delay : int
        Minimum delay between events in milliseconds. Reduces lag
        by limiting the Limit update rate.
    **kwargs :
        Passed to first call of `SliceViewer.update`.
    """
    def __init__(
        self,
        data: np.ndarray,
        update_delay: int = 50,
        **kwargs,
    ):
        self.fig, self.ax = plt.subplots()
        self.data = data
        self.update_delay = update_delay / 1000

        self.max_vals = dict(zip('zyx', np.array(data.shape) - 1))
        self.labels = {
            'x': ('y', 'z'),
            'y': ('x', 'z'),
            'z': ('x', 'y'),
        }

        self.last_update = 0.0

        # Enable direct specification of slice, i.e. x=123
        for along in 'xyz':
            if along in kwargs:
                kwargs['along'] = along
                kwargs['index'] = kwargs[along]
                break

        along = kwargs.get('along', 'x')
        init_max_val = self.max_vals[along]
        init_val = kwargs.get('index', int(init_max_val / 2))

        self.int_slider = IntSlider(value=init_val, min=0, max=init_max_val)
        self.radio_buttons = RadioButtons(options=('x', 'y', 'z'), value=along)

        self.im = self.ax.imshow(data[0], interpolation=None)
        self.im.set_clim(vmin=data.min(), vmax=data.max())
        self.update(index=init_val, along=along)

    def get_slice(self, *, index: int, along: str):
        """Get slice associated with index along given axes."""
        if along == 'x':
            return self.data[:, :, index]
        elif along == 'y':
            return self.data[:, index, :]
        elif along == 'z':
            return self.data[index, ...]
        else:
            raise ValueError('`along` must be one of `x`,`y`,`z`')

    def update(self, index: int, along: str):
        """Update the image in place."""
        now = time.time()
        diff = now - self.last_update

        if diff < self.update_delay:
            return

        max_val = self.max_vals[along]
        xlabel, ylabel = self.labels[along]
        index = min(index, max_val)
        self.int_slider.max = max_val

        slice = self.get_slice(along=along, index=index)

        top, right = slice.shape

        self.im.set_data(slice)
        self.im.set_extent((0, right, 0, top))

        self.ax.set_title(f'slice {index} along {along}')
        self.ax.set_xlabel(xlabel)
        self.ax.set_ylabel(ylabel)
        self.fig.canvas.draw()

        self.last_update = time.time()

    def interact(self):
        """Call interactive `ipywidgets` widget."""
        interact(self.update, index=self.int_slider, along=self.radio_buttons)


def show_image(image,
               *,
               dpi: int = 80,
               title: str = None,
               **kwargs) -> 'plt.Axes':
    """Simple function to show an image using matplotlib.

    Parameters
    ----------
    image : 2D np.ndarray
        Image to display.
    dpi : int, optional
        DPI to render at.
    title : str, optional
        Title for the plot.
    **kwargs : dict
        Extra keyword arguments to pass to `plt.imshow`.

    Returns
    -------
    plt.Axes
        Description
    """
    kwargs.setdefault('interpolation', None)

    fig = plt.figure(dpi=dpi)
    plt.set_cmap('gray')

    ax = fig.add_subplot()
    ax.imshow(image, **kwargs)

    if title:
        plt.title(title)

    ax.set_xlabel('x')
    ax.set_ylabel('y')

    return ax


# https://docs.python.org/3.8/library/itertools.html#itertools-recipes
def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)



from .helpers import compare_mesh_with_image, pad, simple_triangulate
from .mesher import Mesher2D, generate_2d_mesh

__all__ = [
    'compare_mesh_with_image',
    'generate_2d_mesh',
    'Mesher2D',
    'pad',
    'simple_triangulate',
]



from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib.pyplot as plt

if TYPE_CHECKING:
    from nanomesh.mesh_container import MeshContainer


def _legend_with_triplot_fix(ax: plt.Axes, **kwargs):
    """Add legend for triplot with fix that avoids duplicate labels.

    Parameters
    ----------
    ax : plt.Axes
        Matplotlib axes to apply legend to.
    **kwargs
        Extra keyword arguments passed to `ax.legend`.
    """
    handles, labels = ax.get_legend_handles_labels()
    # reverse to avoid blank line color
    by_label = dict(zip(reversed(labels), reversed(handles)))
    ax.legend(by_label.values(), by_label.keys(), **kwargs)


def plot_line_triangle(mesh: MeshContainer, label: str, **kwargs):
    """Plot line/triangle mesh together."""
    assert set(mesh.cells_dict.keys()) == {'line', 'triangle'}
    fig, (ax1, ax2) = plt.subplots(ncols=2)
    mesh.plot_mpl('line', ax=ax1, label=label, **kwargs)
    mesh.plot_mpl('triangle', ax=ax2, label=label, **kwargs)
    return fig, (ax1, ax2)



import os
from ast import literal_eval
from pathlib import Path

import numpy as np


def read_info(filename: os.PathLike) -> dict:
    """Read volume metadata.

    Parameters
    ----------
    filename : PathLike
        Path to the file.

    Returns
    -------
    dct : dict
        Dictionary with the metadata.
    """
    dct = {}
    with open(filename, 'r') as f:
        for line in f:
            line = line.strip()

            if not line:
                continue

            if line.startswith('!'):
                continue

            key, val = line.split('=')

            key = key.strip()
            val = val.strip()

            try:
                val = literal_eval(val)
            except ValueError:
                pass

            dct[key] = val

    return dct


def load_vol(filename: os.PathLike,
             dtype=np.float32,
             mmap_mode: str = None,
             shape: tuple = None) -> np.ndarray:
    """Load data from `.vol` file.

    The image shape is deduced from the `.vol.info` file. If this file is
    not present, the shape can be specified using the `shape` keyword.

    Parameters
    ----------
    filename : os.PathLike
        Path to the file.
    dtype : dtype, optional
        Numpy dtype of the data.
    mmap_mode : None, optional
        If not None, open the file using memory mapping. For more info on
        the modes, see:
        https://numpy.org/doc/stable/reference/generated/numpy.memmap.html
    shape : tuple, optional
        Tuple of three ints specifying the shape of the data (order: z, y, x).

    Returns
    -------
    result : np.ndarray
        Data stored in the file.
    """
    filename = Path(filename)

    if not filename.exists():
        raise IOError(f'No such file: {filename}')

    try:
        filename_info = filename.with_suffix(filename.suffix + '.info')
        if not shape:
            info = read_info(filename_info)
            shape = info['NUM_Z'], info['NUM_Y'], info['NUM_X']
    except FileNotFoundError:
        raise ValueError(
            f'Info file not found: {filename_info.name}, specify '
            'the volume shape using the `shape` parameter.') from None

    result: np.ndarray

    if mmap_mode:
        result = np.memmap(filename, dtype=dtype, shape=shape, mode=mmap_mode)
    else:
        result = np.fromfile(filename, dtype=dtype)
        result = result.reshape(shape)

    return result



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Callable, Union

import matplotlib.pyplot as plt
import numpy as np

from .base_image import BaseImage
from .utils import show_image

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .mesh import TriangleMesh


class Plane(BaseImage):
    @classmethod
    def load(cls, filename: os.PathLike, **kwargs) -> 'Plane':
        """Load the data. Supported filetypes: `.npy`.

        Parameters
        ----------
        filename : Pathlike
            Name of the file to load.
        **kwargs : dict
            Extra keyword arguments passed to `np.load`.

        Returns
        -------
        Plane
            Instance of this class.
        """
        array = np.load(filename, **kwargs)
        return cls(array)

    def apply(self, function: Callable, **kwargs):
        """Apply function to `.image` array. Return an instance of `Plane` if
        the result is a 2D image, otherwise return the result of the operation.

        Parameters
        ----------
        function : callable
            Function to apply to `self.image`.
        **kwargs
            Keyword arguments to pass to `function`.

        Returns
        -------
        Plane
            New instance of `Plane`.
        """
        return super().apply(function, **kwargs)

    def show(self,
             *,
             dpi: int = 80,
             title: str = None,
             **kwargs) -> 'plt.Axes':
        """Plot the image using matplotlib.

        Parameters
        ----------
        dpi : int, optional
            DPI to render at.
        title : str, optional
            Set the title of the plot.
        **kwargs : dict
            Extra keyword arguments to pass to `plt.imshow`.

        Returns
        -------
        ax : `matplotlib.axes.Axes`
            Instance of `matplotlib.axes.Axes`
        """
        return show_image(self.image, dpi=dpi, title=title, **kwargs)

    def generate_mesh(self, **kwargs) -> TriangleMesh:
        """Generate mesh from binary (segmented) image.

        Parameters
        ----------
        **kwargs:
            Keyword arguments are passed to `mesh2d.generate_2d_mesh`

        Returns
        -------
        mesh : TriangleMesh
            Description of the mesh.
        """
        from nanomesh.mesh2d import generate_2d_mesh
        return generate_2d_mesh(image=self.image, **kwargs)

    def select_roi(self, from_points: np.ndarray = None):
        """Select region of interest in interactive matplotlib figure.

        Returns
        -------
        roi : `nanomesh.roi2d.ROISelector`
            Region of interest object. Bounding box is stored in `roi.bbox`.
        """
        from nanomesh.roi2d import ROISelector
        ax = self.show(title='Select region of interest')
        if from_points is not None:
            # reverse columns to match image
            from_points = from_points[:, ::-1]
            ax.scatter(*from_points.T)
        roi = ROISelector(ax, snap_to=from_points)
        return roi

    def crop(self, left: int, top: int, right: int, bottom: int) -> 'Plane':
        """Crop image to pixel indices.

        Parameters
        ----------
        left, top, right, bottom : int
            Index of pixel delimiting cropping box.

        Returns
        -------
        Plane
            New instance of `Plane`.
        """
        return Plane(self.image[top:bottom, left:right])

    def crop_to_roi(self, bbox: np.ndarray) -> 'Plane':
        """Crop plane to rectangle defined by bounding box.

        Parameters
        ----------
        bbox : (4,2) np.ndarray
            List of points describing region of interest. The bounding box
            may be rotated.

        Returns
        -------
        `nanomesh.Plane`
            Cropped region as `nanomesh.Plane` object.
        """
        from nanomesh.roi2d import extract_rectangle
        cropped = extract_rectangle(self.image, bbox=bbox)
        return Plane(cropped)

    def compare_with_mesh(self, mesh: TriangleMesh) -> 'plt.Axes':
        """Make a plot comparing the image with the given mesh.

        Parameters
        ----------
        mesh : TriangleMesh
            Mesh to compare the image with.

        Returns
        -------
        ax : matplotlib.Axes
        """
        from nanomesh.mesh2d import compare_mesh_with_image
        return compare_mesh_with_image(image=self.image, mesh=mesh)

    def compare_with_digitized(self, digitized: Union[np.ndarray, 'Plane'],
                               **kwargs) -> 'plt.Axes':
        """Compare image with digitized (segmented) image. Returns a plot with
        the overlay of the digitized image.

        Parameters
        ----------
        digitized : np.ndarray, Plane
            Digitized image of the same dimensions to overlay
        **kwargs : dict
            Extra keyword arguments passed to `skimage.color.label2rgb`.

        Returns
        -------
        ax : plt.Axes
        """
        from skimage.color import label2rgb

        if isinstance(digitized, Plane):
            digitized = digitized.image

        # bg_label=0 is default for scikit-image from 0.19 onwards
        kwargs.setdefault('bg_label', 0)
        image_overlay = label2rgb(digitized, image=self.image, **kwargs)

        fig, ax = plt.subplots()

        ax.imshow(image_overlay, interpolation='none')
        ax.axis('image')
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_title('Image comparison')

        return ax

    def compare_with_other(self, other: Union[np.ndarray, 'Plane'],
                           **kwargs) -> 'plt.Axes':
        """Compare image with other image.

        Parameters
        ----------
        other : np.ndarray, Plane
            Other image of the same dimensions to overlay
        **kwargs : dict
            Extra keyword arguments passed to `skimage.util.compare_images`.

        Returns
        -------
        ax : plt.Axes
        """
        from skimage.util import compare_images

        if isinstance(other, Plane):
            other = other.image

        kwargs.setdefault('method', 'checkerboard')
        kwargs.setdefault('n_tiles', (4, 4))
        comp = compare_images(self.image, other, **kwargs)

        fig, ax = plt.subplots()

        ax.imshow(comp, interpolation='none')
        ax.axis('image')
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_title(f'Image comparison ({kwargs["method"]})')

        return ax

    def clear_border(self, *, object_label: int, fill_val: int,
                     **kwargs) -> 'Plane':
        """Remove objects at the border of the image.

        Parameters
        ----------
        object_label : int
            Label of the objects to remove.
        fill_val : int
            Cleared objects are set to this value.
        **kwargs
            Extra arguments passed to `skimage.segmentation.clear_border`.

        Returns
        -------
        Plane
            New instance of `Plane`.
        """
        from skimage import segmentation

        objects = (self.image == object_label).astype(int)
        border_cleared = segmentation.clear_border(objects, **kwargs)
        mask = (border_cleared != objects)

        out = self.image.copy()
        out[mask] = fill_val
        return self.__class__(out)

    def try_all_threshold(self, **kwargs):
        """Produce a plot trying all available thresholds from `scikit-image`.

        Parameters
        ----------
        **kwargs : dict
            Extra keyword arguments passed to
            `skimage.filters.try_all_threshold`.
        """
        from skimage import filters
        kwargs.setdefault('verbose', False)
        self.apply(filters.try_all_threshold, **kwargs)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List, Union

import matplotlib.pyplot as plt
import meshio
import numpy as np
from skimage import measure, morphology

from nanomesh._mesh_shared import BaseMesher
from nanomesh.mesh2d import simple_triangulate
from nanomesh.volume import Volume

from ..region_markers import RegionMarker, RegionMarkerLike
from .bounding_box import BoundingBox
from .helpers import pad

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from nanomesh.mesh import TriangleMesh


def get_point_in_prop(
        prop: measure._regionprops.RegionProperties) -> np.ndarray:
    """Uses `skeletonize` to find a point in the center of the regionprop.

    Parameters
    ----------
    prop : RegionProperties
        RegionProp from skimage.measure.regionproperties.

    Returns
    -------
    point : (3,) np.array
        Returns 3 indices describing a pixel in the labeled region.
    """
    skeleton = morphology.skeletonize(prop.image)
    coords = np.argwhere(skeleton)
    middle = len(coords) // 2
    try:
        point = coords[middle]
        point += np.array(prop.bbox[0:3])  # Add prop offset
    except IndexError:
        point = np.array(prop.centroid)
    return point


def get_region_markers(vol: Union[Volume, np.ndarray]) -> List[RegionMarker]:
    """Get region markers describing the featuers in the volume.

    The array will be labeled, and points inside the labeled region
    will be obtained using the `skeletonize` function. The region
    markers can be used to flood the connected regions in the
    tetrahedralization step.

    Parameters
    ----------
    vol : Union[Volume, np.array]
        Segmented integer volume.

    Returns
    -------
    region_markers : List[tuple]
        List of tuples. The first element is the label in the source image,
        and the second the pixel coordinates somewhere in the center of the
        corresponding region.
    """
    region_markers = []

    if isinstance(vol, Volume):
        image = vol.image
    else:
        image = vol

    labels = measure.label(image, background=-1, connectivity=1)

    props = measure.regionprops(labels, intensity_image=image)

    for prop in props:
        point = get_point_in_prop(prop)
        i, j, k = point.astype(int)
        label = image[i, j, k]

        region_marker = RegionMarker(label=label, coordinates=point)
        region_markers.append(region_marker)

    return region_markers


def add_corner_points(mesh: TriangleMesh, bbox: BoundingBox) -> None:
    """Add corner points from bounding box to mesh points.

    Parameters
    ----------
    mesh : TriangleMesh
        Mesh to add corner points to.
    bbox : BoundingBox
        Container for the bounding box coordinates.
    """
    corners = bbox.to_points()
    mesh.points = np.vstack([mesh.points, corners])


def close_side(mesh: TriangleMesh,
               *,
               side: str,
               bbox: BoundingBox,
               ax: plt.Axes = None):
    """Fill a side of the bounding box with triangles.

    Parameters
    ----------
    mesh : TriangleMesh
        Input contour mesh.
    side : str
        Side of the volume to close. Must be one of
        `left`, `right`, `top`, `bottom`, `front`, `back`.
    bbox : BoundingBox
        Coordinates of the bounding box.
    ax : plt.Axes, optional
        Plot the generated side on a matplotlib axis.

    Returns
    -------
    mesh : TriangleMesh
        Triangle mesh with the given side closed.

    Raises
    ------
    ValueError
        When the value of `side` is invalid.
    """
    from nanomesh.mesh import TriangleMesh
    all_points = mesh.points

    if side == 'top':
        edge_col = 2
        edge_value = bbox.zmin
    elif side == 'bottom':
        edge_col = 2
        edge_value = bbox.zmax
    elif side == 'left':
        edge_col = 1
        edge_value = bbox.ymin
    elif side == 'right':
        edge_col = 1
        edge_value = bbox.ymax
    elif side == 'front':
        edge_col = 0
        edge_value = bbox.xmin
    elif side == 'back':
        edge_col = 0
        edge_value = bbox.xmax
    else:
        raise ValueError('Side must be one of `right`, `left`, `bottom`'
                         f'`top`, `front`, `back`. Got {side=}')

    keep_cols = [col for col in (0, 1, 2) if col != edge_col]
    is_edge = all_points[:, edge_col] == edge_value

    coords = all_points[is_edge][:, keep_cols]

    edge_mesh = simple_triangulate(points=coords, opts='')
    cells = edge_mesh.cells_dict['triangle'].copy()

    shape = cells.shape
    new_cells = cells.ravel()

    mesh_edge_index = np.argwhere(is_edge).flatten()
    new_edge_index = np.arange(len(mesh_edge_index))
    mapping = np.vstack([new_edge_index, mesh_edge_index])

    mask = np.in1d(new_cells, mapping[0, :])
    new_cells[mask] = mapping[1,
                              np.searchsorted(mapping[0, :], new_cells[mask])]
    new_cells = new_cells.reshape(shape)

    new_labels = np.ones(len(new_cells))

    points = all_points
    cells = np.vstack([mesh.cells, new_cells])
    labels = np.hstack([mesh.labels, new_labels])

    mesh = TriangleMesh(points=points, cells=cells, labels=labels)

    if ax:
        edge_mesh.plot(ax=ax)
        ax.set_title(side)

    return mesh


def generate_envelope(mesh: TriangleMesh,
                      *,
                      bbox: BoundingBox,
                      plot: bool = False) -> TriangleMesh:
    """Wrap the surface mesh and close any open contours along the bounding
    box.

    Parameters
    ----------
    mesh : TriangleMesh
        Input mesh.
    bbox : BoundingBox
        Coordinates of the bounding box.

    Returns
    -------
    TriangleMesh
        Ouput mesh.
    """
    add_corner_points(mesh, bbox=bbox)

    sides = 'top', 'bottom', 'left', 'right', 'front', 'back'

    for i, side in enumerate(sides):
        mesh = close_side(mesh, side=side, bbox=bbox)

    return mesh


class Mesher3D(BaseMesher):
    def __init__(self, image: np.ndarray):
        super().__init__(image)
        self.contour: TriangleMesh
        self.pad_width = 0

    def generate_contour(
        self,
        level: float = None,
    ):
        """Generate contours using marching cubes algorithm.

        Also generates an envelope around the entire data volume
        corresponding to the bounding box.

        The bounding box equals the dimensions of the data volume.

        Parameters
        ----------
        level : float, optional
            Contour value to search for isosurfaces (i.e. the threshold value).
            By default takes the average of the min and max value. Can be
            ignored if a binary image is passed to `Mesher3D`.
        """
        from nanomesh.mesh import TriangleMesh

        points, cells, *_ = measure.marching_cubes(
            self.image,
            level=level,
            allow_degenerate=False,
        )

        mesh = TriangleMesh(points=points, cells=cells)

        bbox = BoundingBox.from_shape(self.image.shape)
        mesh = generate_envelope(mesh, bbox=bbox)

        logger.info(f'Generated contour with {len(mesh.cells)} cells')

        self.contour = mesh

    def pad_contour(self, **kwargs):
        """Pad the contour. See `.helpers.pad` for info.

        Parameters
        ----------
        **kwargs
            Keyword arguments for `.helpers.pad`.
        """
        self.contour = pad(self.contour, **kwargs)

    def show_contour(self, **kwargs):
        """Pad the contour. See `nanomesh.BaseMesh.plot_pyvista` for info.

        Parameters
        ----------
        **kwargs
            Keyword arguments for `nanomesh.BaseMesh.plot_pyvista`.
        """
        self.contour.plot_pyvista(**kwargs)

    def set_region_markers(self, region_markers: List[RegionMarkerLike]):
        """Sets custom region markers for tetrahedralization.

        Parameters
        ----------
        region_markers : List[RegionMarkerLike]
            List of `RegionMarker` objects or `(int, np.ndarray)` tuples.
        """
        self.contour.region_markers.clear()

        for region_marker in region_markers:
            self.contour.add_region_marker(region_marker)

    def tetrahedralize(self, generate_region_markers: bool = False, **kwargs):
        """Tetrahedralize a surface contour mesh.

        Parameters
        ----------
        generate_region_markers : bool, optional
            Attempt to automatically generate region markers.
            Overwrites existing region_markers.
        **kwargs
            Keyword arguments passed to
            `nanomesh.mesh_container.TriangleMesh.tetrahedralize`

        Returns
        -------
        TetraMesh

        Raises
        ------
        ValueError
            Description
        """
        if not self.contour:
            raise ValueError('No contour mesh available.'
                             'Run `Mesher3D.generate_contour()` first.')

        if generate_region_markers:
            region_markers = get_region_markers(self.image)
            self.contour.region_markers = region_markers

        contour = self.contour
        volume_mesh = contour.tetrahedralize(**kwargs)
        return volume_mesh


def generate_3d_mesh(
    image: np.ndarray,
    *,
    level: float = None,
    **kwargs,
) -> 'meshio.Mesh':
    """Generate mesh from binary (segmented) image.

    Parameters
    ----------
    image : 3D np.ndarray
        Input image to mesh.
    level : float, optional
        Contour value to search for isosurfaces (i.e. the threshold value).
        By default takes the average of the min and max value. Can be
        ignored if a binary image is passed as `image`.
    **kwargs
        Optional keyword arguments passed to
        `nanomesh.mesh_container.TriangleMesh.tetrahedralize`

    Returns
    -------
    volume_mesh : TetraMesh
        Description of the mesh.
    """
    mesher = Mesher3D(image)
    mesher.generate_contour(level=level)

    volume_mesh = mesher.tetrahedralize(**kwargs)
    return volume_mesh



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List, Tuple

import matplotlib.pyplot as plt
import meshio
import numpy as np
from scipy.spatial.distance import cdist
from skimage import measure

from .._mesh_shared import BaseMesher
from .helpers import pad
from .polygon import Polygon

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ..mesh import LineMesh
    from ..mesh_container import MeshContainer


def generate_mesh(polygons: List[Polygon], bbox: np.ndarray) -> LineMesh:
    """Generate line-mesh from polygons and surrounding bbox. The polygons are
    stacked and missing corners are obtained from the bounding box coordinates.

    Parameters
    ----------
    polygons : List[Polygon]
        List of polygons.
    bbox : (n, 2) np.ndarray
        Coordinates for the bounding box. These define the convex hull
        of the meshing area.

    Returns
    -------
    points : (m,2) np.ndarray
        List of points.
    segments : (n,2) np.ndarray
        List of segments.
    """
    from nanomesh import LineMesh

    segments = generate_segments(polygons)

    all_points = np.vstack([polygon.points for polygon in polygons])

    corner_idx = np.argwhere(cdist(bbox, all_points) == 0)

    if len(corner_idx) < len(bbox):
        # Add missing corners and add them where necessary
        missing_corners = np.delete(bbox, corner_idx[:, 0], axis=0)
        all_points = np.vstack([all_points, missing_corners])
        corner_idx = np.argwhere(cdist(bbox, all_points) == 0)

    R = corner_idx[:, 1].tolist()
    additional_segments = list(zip(R, R[1:] + R[:1]))
    segments = np.vstack([segments, additional_segments])

    mesh = LineMesh(points=all_points, cells=segments)
    return mesh


def generate_regions(polygons: List[Polygon]) -> List[Tuple[int, np.ndarray]]:
    """Generate regions for triangle.

    Parameters
    ----------
    polygons : List[Polygon]
        List of polygons.

    Returns
    -------
    regions : (n,5) np.ndarray
        Array with regions with each row: (x, y, z, index, 0)
    """
    regions = []

    for label, polygon in enumerate(polygons):
        point = polygon.find_point()

        # in LineMesh format
        regions.append((1, point))

    return regions


def generate_segments(polygons: List[Polygon]) -> np.ndarray:
    """Generate segments for triangle.

    Parameters
    ----------
    polygons : List[Polygon]
        List of polygons

    Returns
    -------
    segments : np.ndarray
        Segment connectivity array
    """
    i = 0
    segments = []

    for polygon in polygons:
        n_points = len(polygon)
        rng = np.arange(i, i + n_points)

        # generate segment connectivity matrix
        segment = np.vstack([rng, np.roll(rng, shift=-1)]).T
        segments.append(segment)

        i += n_points

    return np.vstack(segments)


class Mesher2D(BaseMesher):
    def __init__(self, image: np.ndarray):
        super().__init__(image)
        self.contour: LineMesh

    def generate_contour(
        self,
        level: float = None,
        contour_precision: int = 1,
        max_contour_dist: int = 5,
    ):
        """Generate contours using marching cubes algorithm.

        Contours are approximated by a polygon, where the maximum distance
        between points is decided by `max_contour_dist`.

        Parameters
        ----------
        level : float, optional
            Contour value to search for isosurfaces (i.e. the threshold value).
            By default takes the average of the min and max value. Can be
            ignored if a binary image is passed to `Mesher2D`.
        contour_precision : int, optional
            Maximum distance from original points in polygon approximation
            routine.
        max_contour_dist : int, optional
            Divide long edges so that maximum distance between points does not
            exceed this value.
        """
        polygons = [
            Polygon(points)
            for points in measure.find_contours(self.image, level=level)
        ]
        polygons = [
            polygon.approximate(contour_precision) for polygon in polygons
        ]
        polygons = [
            polygon.subdivide(max_dist=max_contour_dist)
            for polygon in polygons
        ]
        polygons = [
            polygon.close_corner(self.image.shape) for polygon in polygons
        ]
        polygons = [polygon.remove_duplicate_points() for polygon in polygons]

        regions = generate_regions(polygons)

        contour = generate_mesh(polygons, self.image_bbox)
        contour.add_region_markers(regions)

        self.polygons = polygons
        self.contour = contour

    @property
    def image_bbox(self) -> np.ndarray:
        """Return bbox from image shape.

        Returns
        -------
        bbox : (4,2) np.array
            Coordinates of bounding box contour.
        """
        x, y = self.image.shape
        return np.array((
            (0, 0),
            (x - 1, 0),
            (x - 1, y - 1),
            (0, y - 1),
        ))

    def triangulate(self,
                    opts='pAq30a100',
                    clip_line_data: bool = True,
                    **kwargs) -> MeshContainer:
        """Triangulate contours.

        Mandatory switches for `opts`:
            - e: ensure edges get returned
            - p: planar straight line graph
            - A: assign regional attribute to each triangle

        If missing, these will be added.

        Parameters
        ----------
        opts : str, optional
            Options passed to `triangle.triangulate`
        clip_line_data: bool
            If set, clips the line data to 0: body,
            1: external boundary, 2: internal boundary
            instead of individual numbers for each segment
        **kwargs
            Keyword arguments passed to `triangle.triangulate`

        Returns
        -------
        mesh : MeshContainer
            Output 2D mesh with domain labels
        """
        for var in 'pAe':
            if var not in opts:
                opts = f'{opts}{var}'
        kwargs['opts'] = opts

        mesh = self.contour.triangulate(**kwargs)

        mesh.set_field_data('triangle', {0: 'background', 1: 'feature'})
        mesh.set_field_data('line', {0: 'body', 1: 'external', 2: 'internal'})

        fields = {
            m.label: m.name
            for m in self.contour.region_markers if m.name
        }
        mesh.set_field_data('triangle', fields)

        return mesh

    def generate_domain_mask_from_contours(
        self,
        mesh: MeshContainer,
    ) -> np.ndarray:
        """Generate domain mask from contour.

        Parameters
        ----------
        mesh : MeshContainer
            Input mesh

        Returns
        -------
        labels : (n,) np.array
            Array cell labels.
        """
        centers = mesh.get('triangle').cell_centers

        labels = np.zeros(len(centers), dtype=int)

        for polygon in self.polygons:
            mask = polygon.contains_points(centers)
            labels[mask] = 1

        return labels

    def pad_contour(self, **kwargs):
        """Pad the contour. See `.helpers.pad` for info.

        Parameters
        ----------
        **kwargs
            Keyword arguments for `.helpers.pad`.
        """
        self.contour = pad(self.contour, **kwargs)

    def show_contour(self, ax: plt.Axes = None):
        """Plot contours on image.

        Parameters
        ----------
        ax : matplotlib.Axes
            Axes to use for plotting.

        Returns
        -------
        ax : matplotlib.Axes
        """
        if not ax:
            fig, ax = plt.subplots()

        ax.set_title('Contours')
        self.contour.plot_mpl(ax=ax)

        ax.imshow(self.image)
        ax.axis('image')
        ax.set_xticks([])
        ax.set_yticks([])

        return ax


def generate_2d_mesh(image: np.ndarray,
                     *,
                     level: float = None,
                     max_contour_dist: int = 5,
                     opts: str = 'q30a100',
                     plot: bool = False) -> 'meshio.Mesh':
    """Generate mesh from binary (segmented) image.

    Parameters
    ----------
    image : 2D np.ndarray
        Input image to mesh.
    level : float, optional
        Level to generate contours at from image
    max_contour_dist : int, optional
        Maximum distance between neighbouring pixels in contours.
    opts : str, optional
        Options passed to `triangle.triangulate`

    Returns
    -------
    meshio.Mesh
        Description of the mesh.
    """
    mesher = Mesher2D(image)
    mesher.generate_contour(max_contour_dist=5, level=level)
    return mesher.triangulate(opts=opts)



from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
import triangle as tr
from scipy.spatial.distance import cdist

from ..region_markers import RegionMarker

if TYPE_CHECKING:
    from nanomesh.mesh import LineMesh, TriangleMesh
    from nanomesh.mesh_container import MeshContainer


def compare_mesh_with_image(image: np.ndarray, mesh: TriangleMesh):
    """Compare mesh with image.

    Parameters
    ----------
    image : 2D array
        Image to compare mesh with
    mesh : TriangleMesh
        Triangle mesh to plot on image

    Returns
    -------
    ax : matplotlib.Axes
    """
    fig, ax = plt.subplots()

    mesh.plot_mpl(ax=ax)

    ax.imshow(image)
    ax.axis('image')
    ax.set_xticks([])
    ax.set_yticks([])

    return ax


def simple_triangulate(points: np.ndarray,
                       *,
                       segments: np.ndarray = None,
                       regions: np.ndarray = None,
                       opts: str = '') -> MeshContainer:
    """Simple triangulation using `triangle`.

    Parameters
    ----------
    points : i,2 np.ndarray
        Vertex coordinates.
    segments : j,2 np.ndarray, optional
        Index array describing segments.
        Segments are edges whose presence in the triangulation
        is enforced (although each segment may be subdivided into smaller
        edges). Each segment is specified by listing the indices of its
        two endpoints. A closed set of segments describes a contour.
    regions : k,2 np.ndarray, optional
        Coordinates describing regions. A region is a coordinate inside
        (e.g. at the center) of a region/contour (i.e. enclosed by segments).
    opts : str, optional
        Additional options passed to `triangle.triangulate` documented here:
        https://rufat.be/triangle/API.html#triangle.triangulate

    Returns
    -------
    mesh : MeshContainer
        Triangle mesh
    """
    from nanomesh.mesh_container import MeshContainer

    triangle_dict_in = {'vertices': points}

    if segments is not None:
        triangle_dict_in['segments'] = segments

    if regions is not None:
        triangle_dict_in['regions'] = regions

    triangle_dict_out = tr.triangulate(triangle_dict_in, opts=opts)

    mesh = MeshContainer.from_triangle_dict(triangle_dict_out)

    return mesh


def pad(
    mesh: LineMesh,
    *,
    side: str,
    width: int,
    label: int = None,
    name: str = None,
) -> LineMesh:
    """Pad a triangle mesh (2D).

    Parameters
    ----------
    mesh : LineMesh
        The mesh to pad.
    side : str
        Side to pad, must be one of `left`, `right`, `top`, `bottom`.
    width : int
        Width of the padded area.
    label : int, optional
        The label to assign to the padded area. If not defined, generates the
        next unique label based on the existing ones.
    name : str, optional
        Name of the added region. Note that in case of conflicts, the `label`
        takes presedence over the `name`.

    Returns
    -------
    new_mesh : LineMesh
        Padded line mesh.

    Raises
    ------
    ValueError
        When the value of `side` is invalid.
    """
    labels = [m.label for m in mesh.region_markers]
    names = [m.name for m in mesh.region_markers]

    if (label in labels) and (name is None):
        name = [m.name for m in mesh.region_markers if m.label == label][0]

    if name and (name in names) and (label is None):
        label = [m.label for m in mesh.region_markers if m.name == name][0]

    if label is None:
        label = max(max(labels) + 1, 2)

    if width == 0:
        return mesh

    bottom_edge, right_edge = mesh.points.max(axis=0)
    top_edge, left_edge = mesh.points.min(axis=0)

    if side == 'top':
        corners = np.array([
            [top_edge, right_edge],
            [top_edge - width, right_edge],
            [top_edge - width, left_edge],
            [top_edge, left_edge],
        ])
    elif side == 'left':
        corners = np.array([
            [top_edge, left_edge],
            [top_edge, left_edge - width],
            [bottom_edge, left_edge - width],
            [bottom_edge, left_edge],
        ])
    elif side == 'bottom':
        corners = np.array([
            [bottom_edge, right_edge],
            [bottom_edge + width, right_edge],
            [bottom_edge + width, left_edge],
            [bottom_edge, left_edge],
        ])
    elif side == 'right':
        corners = np.array([
            [top_edge, right_edge],
            [top_edge, right_edge + width],
            [bottom_edge, right_edge + width],
            [bottom_edge, right_edge],
        ])
    else:
        raise ValueError('Side must be one of `right`, `left`, `bottom`'
                         f'`top`. Got {side=}')

    all_points = mesh.points

    corner_idx = np.argwhere(cdist(corners, all_points) == 0)

    if len(corner_idx) < len(corners):
        # Add missing corners and add them where necessary
        missing_corners = np.delete(corners, corner_idx[:, 0], axis=0)
        all_points = np.vstack([all_points, missing_corners])
        corner_idx = np.argwhere(cdist(corners, all_points) == 0)

    R = corner_idx[:, 1].tolist()
    additional_segments = list(zip(R, R[1:] + R[:1]))
    cells = np.vstack([mesh.cells, additional_segments])

    center = corners.mean(axis=0)
    region_markers = mesh.region_markers + [RegionMarker(label, center, name)]

    new_mesh = mesh.__class__(
        points=all_points,
        cells=cells,
        region_markers=region_markers,
    )

    return new_mesh



from dataclasses import dataclass
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
from skimage import measure


@dataclass
class Polygon:
    # List of coordinates describing a polygon.
    points: np.ndarray

    def __len__(self):
        return len(self.points)

    def find_point(self) -> np.ndarray:
        """Use rejection sampling to find point in polygon.

        Returns
        -------
        point : np.ndarray
            Coordinate of point in the polygon
        """
        # start with guess in center of polygon
        point = self.points.mean(axis=0)

        while not measure.points_in_poly([point], self.points):
            xmin, ymin = self.points.min(axis=0)
            xmax, ymax = self.points.max(axis=0)
            point = np.random.uniform(xmin,
                                      xmax), np.random.uniform(ymin, ymax)

        return point

    def close_corner(self, shape: tuple) -> 'Polygon':
        """Check if polygons are in the corner, and close them if needed.

        Polygons which cover a corner cannot be closed by joining the first
        and last element, because some of the area is missed. This algorithm
        adds the corner point to close the polygons.

        Parameters
        ----------
        shape : tuple
            Shape of the source image. Used to check which corners the
            polygon touches.

        Returns
        -------
        polygon : Polygon
            Return a polygon with a corner point added if needed (n+1,2),
            otherwise return the input polygon (n,2)
        """
        xmin, ymin = self.points.min(axis=0)
        xmax, ymax = self.points.max(axis=0)

        xdim, ydim = np.array(shape) - 1

        left = (xmin == 0)
        right = (xmax == xdim)
        bottom = (ymin == 0)
        top = (ymax == ydim)

        if bottom and left:
            extra_point = (0, 0)
        elif top and left:
            extra_point = (0, ydim)
        elif top and right:
            extra_point = (xdim, ydim)
        elif bottom and right:
            extra_point = (xdim, 0)
        else:
            # all good
            return self

        points = np.vstack([self.points, extra_point])
        return Polygon(points)

    def subdivide(self, max_dist: int = 10, plot: bool = False) -> 'Polygon':
        """This algorithm looks for long edges in the polygon and subdivides
        them so they are no longer than `max_dist`

        Parameters
        ----------
        max_dist : int, optional
            Maximum distance between neighbouring coordinates.
        plot : bool, optional
            Show plot of the generated points.

        Returns
        -------
        Polygon
            Polygon with updated coordinate array.
        """
        points = self.points

        new_points: Any = []
        rolled = np.roll(points, shift=-1, axis=0)
        diffs = rolled - points
        # ignore last point, do not wrap around
        dist = np.linalg.norm(diffs[:-1], axis=1)

        last_i = 0

        for i in np.argwhere(dist > max_dist).reshape(-1, ):
            new_points.append(points[last_i:i])
            start = points[i]
            stop = rolled[i]
            to_add = int(dist[i] // max_dist)
            interpolated = np.linspace(start, stop, to_add, endpoint=False)
            new_points.append(interpolated)

            last_i = i + 1

        new_points.append(points[last_i:])
        new_points = np.vstack(new_points)

        if plot:
            plt.scatter(*points.T[::-1], color='red', s=100, marker='x')
            plt.plot(*points.T[::-1], color='red')
            plt.scatter(*new_points.T[::-1], color='green', s=100, marker='+')
            plt.plot(*new_points.T[::-1], color='green')
            plt.axis('equal')
            plt.show()

        return Polygon(new_points)

    def remove_duplicate_points(self) -> 'Polygon':
        """Remove duplicate points from polygon.

        For a polygon it is implied that the last point connects to the
        first point. In case the first point equals the last point, this
        results in errors down the line.

        Returns
        -------
        Polygon
        """
        points = self.points

        first = points[0]
        last = points[-1]

        if np.all(first == last):
            points = points[:-1]

        return Polygon(points)

    def approximate(self, *args, **kwargs) -> 'Polygon':
        """Approximate polygon.

        Parameters
        ----------
        *args : list
            Extra arguments padded to `skimage.measure.approximate_polygon`.
        **kwargs : dict
            Extra keyword arguments passed to
            `skimage.measure.approximate_polygon`.

        Returns
        -------
        new_polygon : Polygon
        """
        new_points = measure.approximate_polygon(self.points, *args, **kwargs)
        return Polygon(new_points)

    def contains_points(self, points: np.ndarray) -> np.ndarray:
        """Test whether points lie inside a polygon.

        Parameters
        ----------
        points : (n,2) np.ndarray
            Description

        Returns
        -------
        mask : (n,) boolean np.ndarray
            True if corresponding point is inside the p olygon
        """
        return measure.points_in_poly(points, self.points)



from .bounding_box import BoundingBox
from .helpers import pad
from .mesher import Mesher3D, generate_3d_mesh, get_region_markers

__all__ = [
    'BoundingBox',
    'generate_3d_mesh',
    'get_region_markers',
    'Mesher3D',
    'pad',
]
